from __future__ import annotations

"""
MIT License

Copyright (c) 2022 Kelly M. Nunez Ocasio

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
"""
Core routines that make up the TAMS algorithm.
"""


import functools
import logging
import warnings
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd  # or just pandas?
import xarray as xr

if TYPE_CHECKING:
    import geopandas
    import matplotlib
    import numpy
    import shapely
    import xarray

###########################################################
###########################################################


def sort_ew(cs: geopandas.GeoDataFrame):
    """Sort the frame east to west descending, using the centroid lon value."""
    # TODO: optional reset_index ?
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            category=UserWarning,
            message="Geometry is in a geographic CRS. Results from 'centroid' are likely incorrect.",
        )
        # fmt: off
        return (
            cs
            .assign(x=cs.geometry.centroid.x)
            .sort_values("x", ascending=False)
            .drop(columns="x")
        )
        # fmt: on


def plot_tracked(
    cs: geopandas.GeoDataFrame,
    *,
    alpha: float = 0.25,
    background: str = "countries",
    label: str = "id",
    add_colorbar: bool = False,
    cbar_kwargs: dict | None = None,
    ax: matplotlib.axes.Axes | None = None,
    size: float = 4,
    aspect: float | None = None,
):
    """Plot CEs at a range of times (colors) with CE group ID (MCS ID) identified.

    Parameters
    ----------
    cs
        Tracked CEs (with MCS ID column).
    alpha
        Alpha applied when plotting the CEs.
    background : {"map", "countries", "none"}
        "map" uses Mercator projection Cartopy's stock image.
        "countries" uses Mercator projection and adds Cartopy coastlines and country borders.
        "none" plots without projection or background.
        This setting is only relevant if `ax` is not provided.
    label : {"id", "none"}
        "id": label each CE with its MCS ID.
        "none": don't label CEs.
    add_colorbar
        Add colorbar with time info.
    cbar_kwargs
        Keyword arguments to pass to ``plt.colorbar``.
    ax
        Axes to plot on. If not provided, a new figure is created.
        The figure size used is ``(size * aspect, size)``.
    size
        Height of the figure (in inches) if `ax` is not provided.
    aspect
        Figure width : height.
        If not provided, it is estimated using the ``total_bounds`` of `cs`.
    """

    import matplotlib.pyplot as plt
    import pandas as pd
    from matplotlib import patheffects

    valid_backgrounds = {"map", "countries", "none"}
    if background not in valid_backgrounds:
        raise ValueError(f"`background` must be one of {valid_backgrounds}")

    valid_labels = {"id", "none"}
    if label not in valid_labels:
        raise ValueError(f"`label` must be one of {valid_labels}")

    if aspect is None:
        # TODO: maybe better to use `.envelope`
        x0, y0, x1, y1 = cs.total_bounds
        aspect = (x1 - x0) / (y1 - y0)
        # ^ estimate for controlling figure size like in xarray
        # https://xarray.pydata.org/en/stable/user-guide/plotting.html#controlling-the-figure-size

    blob_kwargs = dict(alpha=alpha, lw=1.5)
    text_kwargs = dict(
        fontsize=14,
        zorder=10,
        path_effects=[patheffects.withStroke(linewidth=2, foreground="0.2")],
    )
    if ax is None:
        if background in {"map", "countries"}:
            try:
                import cartopy.crs as ccrs
            except ImportError as e:
                raise RuntimeError("cartopy required") from e

            proj = ccrs.Mercator()
            tran = ccrs.PlateCarree()
            fig = plt.figure(figsize=(size * aspect, size))
            ax = fig.add_subplot(projection=proj)
            ax.set_extent([x0, x1, y0, y1])  # type: ignore[attr-defined]
            ax.gridlines(draw_labels=True)  # type: ignore[attr-defined]

            if background == "map":
                # TODO: a more high-res image
                ax.stock_img()  # type: ignore[attr-defined]
            else:  # countries
                import cartopy.feature as cfeature

                ax.add_feature(cfeature.BORDERS, linewidth=0.7, edgecolor="0.3")  # type: ignore[attr-defined]
                ax.coastlines()  # type: ignore[attr-defined]

            blob_kwargs.update(transform=tran)
            text_kwargs.update(transform=tran)

        else:  # none
            _, ax = plt.subplots()

            ax.set(xlabel="lon [°E]", ylabel="lat [°N]")

    t = pd.Series(sorted(cs.time.unique()))
    tmin, tmax = t.iloc[0], t.iloc[-1]
    dt = t.diff().min()

    def get_color(t_):
        if tmin == tmax:
            return plt.cm.tab10.colors[0]
        else:
            x = (t_ - tmin) / (tmax - tmin) * 0.65 + 0.2
            return plt.cm.GnBu(x)

    # Plot blobs at each time
    for t_, g in cs.groupby("time"):
        color = get_color(t_)
        blob_kwargs.update(facecolor=color, edgecolor=color)
        text_kwargs.update(color=color)

        g.plot(ax=ax, **blob_kwargs)

        # Label blobs with assigned ID
        if label == "id":
            with warnings.catch_warnings():
                warnings.filterwarnings(
                    "ignore",
                    category=UserWarning,
                    message="Geometry is in a geographic CRS. Results from 'centroid' are likely incorrect.",
                )
                for id_, x, y in zip(g.mcs_id, g.centroid.x, g.centroid.y):
                    ax.text(x, y, id_, **text_kwargs)  # type: ignore[arg-type]

    if add_colorbar:
        import matplotlib as mpl

        cbar_kwargs_default = {
            "orientation": "horizontal",
            "label": f"hours since {cs.time.min():%Y-%m-%d %H:%M}",
            "shrink": 0.7,
            "aspect": 40,
        }
        if cbar_kwargs is None:
            cbar_kwargs = {}
        cbar_kwargs = {**cbar_kwargs_default, **cbar_kwargs}

        cmap = mpl.colors.ListedColormap([get_color(t_) for t_ in t])
        hours = ((t - tmin).dt.total_seconds() / 3600).values
        hours = np.append(hours, hours[-1] + dt.total_seconds() / 3600)
        m = mpl.cm.ScalarMappable(
            cmap=cmap, norm=mpl.colors.BoundaryNorm(hours, cmap.N)
        )
        plt.colorbar(m, ax=ax, **cbar_kwargs)


def _the_unique(s: pd.Series):
    """Return the one unique value or raise ValueError."""
    u = s.unique()
    if u.size == 1:
        return u[0]
    else:
        raise ValueError(f"the Series has more than one unique value: {u}")


###########################################################
###########################################################

### TRACKING Functions

logger = logging.getLogger(__name__)


def contours(
    x: xarray.DataArray,
    value: float,
    *,
    unstructured: bool | None = None,
    **kwargs,
) -> list[numpy.ndarray]:
    """Find contour definitions for data `x` at value `value`.

    Parameters
    ----------
    x
        Data to be contoured.
        Currently needs to have ``'lat'`` and ``'lon'`` coordinates.
        The array should be 2-D if structured, 1-D if unstructured.
    value
        Find contours where `x` has this value.
    unstructured
        Whether the grid of `x` is unstructured (e.g. MPAS native output).
        Default: assume unstructured if `x` is 1-D.

    Returns
    -------
    :
        List of 2-D arrays describing contours.
        The arrays are shape ``(n, 2)``; each row is a coordinate pair.
    """
    if x.isnull().all():
        raise ValueError("Input array `x` is all null (e.g. NaN)")

    if unstructured is None:
        unstructured = x.ndim == 1

    # TODO: have this return GDF instead?
    # TODO: allowing specifying `crs`, `method`, shapely options (buffer, convex-hull), ...
    # TODO: optionally filter out unclosed?
    import matplotlib.pyplot as plt

    if unstructured:
        assert x.ndim == 1, "this is for a single time step"
        tri = kwargs.get("triangulation")
        if tri is None:
            from matplotlib.tri import Triangulation

            tri = Triangulation(x=x.lon, y=x.lat)
        with plt.ioff():  # requires mpl 3.4
            fig = plt.figure()
            cs = plt.tricontour(tri, x, levels=[value])
    else:
        assert x.ndim == 2, "this is for a single image"
        with plt.ioff():  # requires mpl 3.4
            fig = plt.figure()
            cs = x.plot.contour(x="lon", y="lat", levels=[value])  # type: ignore[attr-defined]
            # cs = x.plot.contour(x=x.dims[2], y=x.dims[1], levels=[value])
            # originally x = 'lon', y = 'lat'

    plt.close(fig)
    assert len(cs.allsegs) == 1, "only one level"

    return cs.allsegs[0]


def _contours_to_gdf(cs: list[np.ndarray]) -> geopandas.GeoDataFrame:
    from geopandas import GeoDataFrame
    from shapely.geometry.polygon import LinearRing, LineString, Point, orient

    polys = []
    for c in cs:
        x, y = c.T
        if x.size < 3:
            logger.info(f"skipping an input contour with less than 3 points: {c}")
            continue
        r = LinearRing(zip(x, y))
        p0 = r.convex_hull  # TODO: optional, also add buffer option
        if type(p0) is LineString or type(p0) is Point:
            continue  # not a closed contour
        p = orient(p0)  # -> counter-clockwise
        polys.append(p)

    return GeoDataFrame(geometry=polys, crs="EPSG:4326")
    # ^ This crs indicates input in degrees


def _size_filter_contours(
    cs235: geopandas.GeoDataFrame,
    cs219: geopandas.GeoDataFrame,
    *,
    threshold: float = 4000,
) -> tuple[geopandas.GeoDataFrame, geopandas.GeoDataFrame]:
    """Compute areas and use to filter both sets of contours.

    `threshold` is for the total cold-core contour area within a given CE contour
    (units: km2).
    """
    import geopandas as gpd
    import numpy as np
    from shapely.geometry import MultiPolygon

    # Drop small 235s (a 235 with area < 4000 km2 can't have 219 area of 4000)
    cs235["area_km2"] = cs235.to_crs("EPSG:32663").area / 10**6

    # if sum(cs235["area_km2"].is_valid) == 0:
    #     cs235["area_km2"] = cs235.to_crs('EPSG:4087').area / 10**6
    #     area = cs235["area_km2"]
    #     print(f'the prior was nan so now it is \n {area}')
    # cs235["area_km2"] = cs235.area / 10**6

    # ^ This crs is equidistant cylindrical
    big_enough = cs235.area_km2 >= threshold
    if not big_enough.empty:
        logger.info(
            f"{big_enough.value_counts().get(True, 0) / big_enough.size * 100:.1f}% "
            f" of 235s are big enough ({threshold} km2)"
        )
    cs235 = cs235[big_enough].reset_index(drop=True)

    # Drop very small 219s (insignificant)
    # Note: This wasn't done in original TAMS, but here we are sometimes seeing
    # tiny 219 contours inside larger ones.
    individual_219_threshold = 10  # km2
    cs219["area_km2"] = cs219.to_crs("EPSG:32663").area / 10**6
    big_enough = cs219.area_km2 >= individual_219_threshold
    if not big_enough.empty:
        logger.info(
            f"{big_enough.value_counts().get(True, 0) / big_enough.size * 100:.1f}% "
            f" of 219s are big enough ({individual_219_threshold} km2)"
        )
    cs219 = cs219[big_enough].reset_index(drop=True)

    # Identify indices of 219s inside 235s
    # Note that some 235s might not have any 219s inside
    a = cs235.sjoin(cs219, predicate="contains", how="left").reset_index()
    # ^ gives an Int64 index with duplicated values, for each 219 inside a certain 235
    i219s = {  # convert to list
        i235: sorted(g.index_right.astype(int).to_list())
        for i235, g in a.groupby("index")
        if not g.index_right.isna().all()
    }

    # Store 219 info inside the 235 df
    cs235["inds219"] = [i219s.get(i235, []) for i235 in cs235.index]
    # TODO: possible to instead cleanly store the 219 `MultiPolygon`s in a column?
    # cs235["cs219"] = cs235.inds219.apply(lambda inds: cs219.iloc[inds][["geometry"]].dissolve().geometry)
    # cs235["cs219"] = cs235.inds219.apply(lambda inds: MultiPolygon(cs219.geometry.iloc[inds].values))

    # Check 219 area sum inside the 235 (total 219 area 4000 km2)
    sum219s = {
        i235: cs219.iloc[i219s.get(i235, [])].to_crs("EPSG:32663").area.sum() / 10**6
        for i235 in cs235.index
    }
    cs235["area219_km2"] = pd.Series(sum219s, dtype=float)
    big_enough = cs235.area219_km2 >= threshold
    if not big_enough.empty:
        logger.info(
            f"{big_enough.value_counts().get(True, 0) / big_enough.size * 100:.1f}% "
            f"of big-enough 235s have enough 219 area ({threshold} km2)"
        )
    cs235 = cs235[big_enough].reset_index(drop=True)

    # Store 219s inside the 235 frame as `MultiPolygon`s
    cs235["cs219"] = gpd.GeoSeries(
        cs235.inds219.apply(lambda inds: MultiPolygon(cs219.geometry.iloc[inds].values))
    )
    # NOTE: above method I found to be ~ 10x faster than this alternative:
    #   gpd.GeoSeries(cs235.inds219.apply(lambda inds: cs219.iloc[inds][["geometry"]].dissolve().geometry[0]))
    # The other benefit of applying `MultiPolygon` directly is that we get `MultiPolygon`
    # even in the case of only one, whereas dissolve gives `Polygon` in that case (row).
    #
    # TODO: alternative method dissolve on the 219s for each row like below leads to
    # slightly different data-in-contours pixel counts (greater?) in some cases
    #   f = lambda row: data_in_contours(tb, cs219.iloc[row.inds219].dissolve())
    #   res_ = cs235.apply(f, axis="columns")  # Series of DataFrames
    #   res = pd.concat(res_.to_list()).reset_index(drop=True)
    # TODO: unary_union (like dissolve uses), might be better unless can properly deal with
    # the small holes we sometimes get, otherwise some area gets counted extra

    # TODO: some elegant way to drop 219s that aren't inside a 235, resetting index
    # but preserving the matching of 235 to 219s

    return cs235, cs219


def _identify_one(
    ctt: xr.DataArray,
    *,
    size_filter: bool = True,
    ctt_threshold: float = 235,
    ctt_core_threshold: float = 219,
    unstructured: bool | None = None,
    triangulation: matplotlib.tri.Triangulation | None = None,
) -> tuple[geopandas.GeoDataFrame, geopandas.GeoDataFrame]:
    """Identify clouds in 2-D cloud-top temperature data `ctt` (e.g. at a specific time)."""

    cs235 = sort_ew(
        _contours_to_gdf(
            contours(
                ctt,
                ctt_threshold,
                unstructured=unstructured,
                triangulation=triangulation,
            )
        )
    ).reset_index(drop=True)
    cs219 = sort_ew(
        _contours_to_gdf(
            contours(
                ctt,
                ctt_core_threshold,
                unstructured=unstructured,
                triangulation=triangulation,
            )
        )
    ).reset_index(drop=True)

    if size_filter:
        cs235, cs219 = _size_filter_contours(cs235, cs219)

    return cs235, cs219


def identify(
    ctt: xarray.DataArray,
    *,
    size_filter: bool = True,
    parallel: bool = False,
    ctt_threshold: float = 235,
    ctt_core_threshold: float = 219,
) -> tuple[list[geopandas.GeoDataFrame], list[geopandas.GeoDataFrame]]:
    """Identify clouds in 2-D (lat/lon) or 3-D (lat/lon + time) cloud-top temperature data `ctt`.
    The 235 K contours returned (first list) serve to identify cloud elements (CEs).
    In a given frame from this list, each row corresponds to a certain CE.

    This is the first step in a TAMS workflow.

    Parameters
    ----------
    ctt
        Cloud-top temperature array.
    size_filter
        Whether to apply size-filtering
        (using 235 K and 219 K areas to filter out CEs that are not MCS material).
        Filtering at this stage makes TAMS more computationally efficient overall.
        Disable this option to return all identified CEs.
        Note that all 219s are returned regardless of this setting.

        When enabled, this also identifies the 219s (if any) that are within each 235.
    parallel
        Identify in parallel along ``'time'`` dimension for 3-D `ctt` (requires `joblib`).
    ctt_threshold
        Used to identify the edges of cloud elements.
    ctt_core_threshold
        Used to identify deep convective cloud regions within larger cloud areas.
        This is used to determine whether or not a system is eligible for being classified
        as an organized system.
        It helps target raining clouds.
    """
    dims = tuple(ctt.dims)

    unstructured = (len(dims) == 2 and "time" in dims) or (len(dims) == 1)

    triangulation = None
    if unstructured:
        if ctt.lat.ndim == 1 and ctt.lon.ndim == 1:
            from matplotlib.tri import Triangulation

            triangulation = Triangulation(x=ctt.lon, y=ctt.lat)
        else:
            warnings.warn(
                "detected unstructured data but not 1-D lat/lon "
                f"(got lat dims {ctt.lat.dims}, lon dims {ctt.lon.dims}), "
                "not pre-computing the triangulation"
            )

    f = functools.partial(
        _identify_one,
        size_filter=size_filter,
        ctt_threshold=ctt_threshold,
        ctt_core_threshold=ctt_core_threshold,
        unstructured=unstructured,
        triangulation=triangulation,
    )

    if (not unstructured and len(dims) == 2) or (unstructured and len(dims) == 1):
        cs235, cs219 = f(ctt)
        css235, css219 = (cs235,), (cs219,)  # to tuple for consistency

    elif "time" in dims and (
        (not unstructured and len(dims) == 3) or (unstructured and len(dims) == 2)
    ):
        assert ctt.time.ndim == 1
        itimes = np.arange(ctt.time.size)

        if parallel:
            try:
                import joblib
            except ImportError as e:
                raise RuntimeError("joblib required") from e

            res = joblib.Parallel(n_jobs=-2, verbose=10)(
                joblib.delayed(f)(ctt.isel(time=i)) for i in itimes
            )

        else:
            res = [f(ctt.isel(time=i)) for i in itimes]

        css235, css219 = zip(*res)

    else:
        raise ValueError(
            f"Got unexpected `ctt` dims: {dims}. "
            "They should be 2-D (lat/lon) + optional 'time' dim for structured grid data, "
            "or 1-D (cell) + optional 'time' dim for unstructured grid data."
        )

    return list(css235), list(css219)


def _data_in_contours_sjoin(
    data: xr.DataArray | xr.Dataset,
    contours: geopandas.GeoDataFrame,
    *,
    varnames: list[str],
    agg=("mean", "std", "count"),
) -> geopandas.GeoDataFrame:
    """Compute stats on `data` within `contours` using :func:`~geopandas.tools.sjoin`.

    `data` must have ``'lat'`` and ``'lon'`` variables.
    """
    import geopandas as gpd

    # Convert possibly-2-D data to GeoDataFrame of points
    print(f"data is {data.dims}")
    data_df = data.to_dataframe().reset_index(drop=set(data.dims) != {"lat", "lon"})
    print(f"data_df is {data_df}")
    lat = data_df["lat"].values
    lon = data_df["lon"].values
    print(f"lat is \n {lat}")
    print(f"\n lon is \n {lon}")
    geom = gpd.points_from_xy(lon, lat, crs="EPSG:4326")  # can be slow with many points
    points = gpd.GeoDataFrame(data_df, geometry=geom)
    print(f"points are {points}")
    print(f"contours is {contours}")

    # Determine which contour (if any) each point is inside
    points = points.sjoin(contours, predicate="within", how="left", rsuffix="contour")
    points = points.dropna().convert_dtypes()
    points["lat"] = points.geometry.y
    points["lon"] = points.geometry.x

    # Aggregate points inside contour
    # TODO: a way to do this without groupby loop?
    new_data_ = {}
    for i, g in points.groupby("index_contour"):
        r = g[varnames].agg(agg).T  # columns: aggs; rows: variables
        new_data_[i] = r
    if not new_data_:
        raise ValueError("no data found in contours")
    new_data = pd.concat(new_data_).convert_dtypes()

    # Convert to standard (non-multi) index and str columns
    new_data = new_data.unstack()  # multi index -> (variable, agg) columns
    new_data.columns = ["_".join(s for s in tup) for tup in new_data.columns]

    return new_data


def _data_in_contours_regionmask(
    data: xr.DataArray | xr.Dataset,
    contours: geopandas.GeoDataFrame,
    *,
    varnames: list[str],
    agg=("mean", "std", "count"),
) -> geopandas.GeoDataFrame:
    import regionmask

    # Form regionmask(s)
    shapes = contours[["geometry"]]
    regions = regionmask.from_geopandas(shapes)
    mask = regions.mask(data)
    # Note: before Shapely v2, having `pygeos` installed made this faster

    # Aggregate points inside contour
    new_data_ = {
        i: data.where(mask == i).to_dataframe()[varnames].dropna().agg(agg).T
        for i in regions.numbers
    }
    if not new_data_:
        raise ValueError("no data found in contours")
    new_data = pd.concat(new_data_).convert_dtypes()
    # TODO: also try with xarray methods instead of going through pandas
    # TODO: try with xarray groupby

    # Convert to standard (non-multi) index and str columns
    new_data = new_data.unstack()  # multi index -> (variable, agg) columns
    new_data.columns = ["_".join(s for s in tup) for tup in new_data.columns]

    return new_data


def data_in_contours(
    data: xarray.DataArray | xarray.Dataset,
    contours: geopandas.GeoDataFrame,
    *,
    agg=("mean", "std", "count"),  # TODO: type
    method: str = "sjoin",
    merge: bool = False,
) -> geopandas.GeoDataFrame:
    """Compute statistics on `data` within the shapes of `contours`.

    With the default settings, we calculate,
    for each shape (row) in the `contours` dataframe:

    - the mean value of `data` within the shape
    - the standard deviation of `data` within the shape
    - the count of non-null values of `data` within the shape

    Parameters
    ----------
    data
        It should have ``'lat'`` and ``'lon'`` coordinates.
    contours
        For example, dataframe of CE or MCS shapes, e.g.
        from :func:`identify` or :func:`track`.
    agg : sequence of str or callable
        Suitable for passing to :meth:`pandas.DataFrame.aggregate`.
    method : {'sjoin', 'regionmask'}
        The regionmask method is suited for data on a structured grid,
        while the GeoPandas sjoin method works for scattered point data as well.
        The sjoin method is the default since it is more general
        and currently often faster.
    merge
        Whether to merge the new data with `contours` or return a separate frame.

    See Also
    --------
    :ref:`ctt_thresh_data_in_contours`
        A usage example.
    """
    if isinstance(data, xr.DataArray):
        varnames = [data.name]
        if data.isnull().all():
            raise ValueError("Input array `data` is all null (e.g. NaN)")
            # TODO: warn instead for this and return cols of NaNs?
    elif isinstance(data, xr.Dataset):
        # varnames = [vn for vn in field.variables if vn not in {"lat", "lon"}]
        raise NotImplementedError
    else:
        raise TypeError

    if isinstance(agg, str):
        agg = (agg,)

    args = (data, contours)
    kwargs = dict(varnames=varnames, agg=agg)

    if method in {"sjoin", "geopandas", "gpd"}:
        new_data = _data_in_contours_sjoin(*args, **kwargs)
    elif method in {"regionmask"}:
        new_data = _data_in_contours_regionmask(*args, **kwargs)
    else:
        raise ValueError(f"method {method!r} not recognized")

    if merge:
        # Merge with the `contours` gdf, appending columns
        new_data = contours.merge(
            new_data, left_index=True, right_index=True, how="left"
        )

    return new_data


def _project_geometry(s: geopandas.GeoSeries, *, dx: float) -> geopandas.GeoSeries:
    crs0 = s.crs.to_string()

    return s.to_crs(crs="EPSG:32663").translate(xoff=dx).to_crs(crs0)
    # return s.to_crs(crs="EPSG:4326").translate(xoff=dx).to_crs(crs0)


# TODO: test


def project(df: geopandas.GeoDataFrame, *, u: float = 0, dt: float = 3600):
    """Project the coordinates by `u` * `dt` meters in the *x* direction.

    Parameters
    ----------
    df
        Dataframe of objects to be spatially projected.
    u
        Speed [m s-1].
    dt
        Time [s]. Default: one hour.
    """
    dx = u * dt
    new_geometry = _project_geometry(df.geometry, dx=dx)

    return df.assign(geometry=new_geometry)


def overlap(a: geopandas.GeoDataFrame, b: geopandas.GeoDataFrame, *, norm: str = "a"):
    """For each contour in `a`, determine those in `b` that overlap and by how much.

    Currently the mapping is based on indices of the frames:
    iloc position in `a` : loc position in `b` : overlap fraction.

    Parameters
    ----------
    norm : {'a', 'b', 'max', 'min', 'mean'}
        Area to use to normalize the overlap to a fraction.
    """
    # TODO: test(s) for the different `norm` options
    s_crs_area = "EPSG:32663"
    # s_crs_area = "EPSG:4326"
    area_a = a.to_crs(s_crs_area).area
    area_b = b.to_crs(s_crs_area).area

    res = {}
    for i in range(len(a)):
        a_i = a.iloc[i : i + 1]  # slicing preserves GeoDataFrame type
        a_i_poly = a_i.values[0][0]
        with warnings.catch_warnings():
            # We get this warning when an empty intersection is found
            warnings.filterwarnings(
                "ignore",
                category=RuntimeWarning,
                message="invalid value encountered in intersection",
            )
            inter = b.intersection(a_i_poly)
        inter = inter[~inter.is_empty]

        ov = inter.to_crs(s_crs_area).area

        if norm == "a":
            area_norm = area_a.iloc[i]
        elif norm == "b":
            area_norm = area_b.loc[ov.index]
        elif norm in {"max", "min", "mean"}:
            op = norm
            b_area_i = area_b.loc[ov.index]
            area_norm = getattr(
                pd.concat(
                    [
                        pd.Series(
                            data=np.full(len(b_area_i), area_a.iloc[i]),
                            index=b_area_i.index,
                        ),
                        b_area_i,
                    ],
                    axis="columns",
                ),
                op,
            )(axis="columns")
        else:
            raise ValueError(f"invalid `norm` {norm!r}")

        ov_frac = ov / area_norm

        res[i] = ov_frac.to_dict()

    return res


def track(
    contours_sets: list[geopandas.GeoDataFrame],
    times,  # TODO: could replace these two with single dict?
    *,
    overlap_threshold: float = 0.5,
    u_projection: float = 0,
    durations=None,
    look: str = "back",
    largest: bool = False,
    overlap_norm: str | None = None,
) -> geopandas.GeoDataFrame:
    """Assign group IDs to the CEs identified at each time, returning a single CE frame.

    Currently this works by: for each CE at the current time step,
    searching for a "parent" from the previous time step by computing
    overlap with all previous CEs.

    Parameters
    ----------
    contours_sets
        List of identified contours, in GeoDataFrame format.
    times
        Timestamps associated with each identified set of contours.
    overlap_threshold
        In [0, 1] (i.e., fractional), the overlap threshold.
    u_projection
        Zonal projection velocity, to project previous time step CEs by before
        computing overlap.
        5--13 m/s are typical magnitudes to use.
        For AEWs, a negative value should be used.
    durations
        Durations associated with the times in `times` (akin to the time resolution).
        If not provided, they will be estimated using ``times[1:] - times[:-1]``.
    look
        (time) direction in which we "look" and compute overlaps,
        linking CEs in time.
    largest
        Only the largest CE continues a track.
    overlap_norm
        Passed to :func:`overlap`.
        Default is to use child area (
        ``'a'`` for ``look='back'``,
        ``'b'`` for ``look='forward'``
        ), i.e. the CE at the later of the two times.
    """
    assert len(contours_sets) == len(times) and len(times) > 1
    times = pd.DatetimeIndex(times)
    itimes = list(range(times.size))

    if durations is not None:
        assert len(durations) == len(times)
    else:
        # Estimate dt values
        dt = times[1:] - times[:-1]
        assert (dt.astype(np.int64) > 0).all()
        if not dt.unique().size == 1:
            warnings.warn("unequal time spacing")
        dt = dt.insert(-1, dt[-1])

    # IDEA: even at initial time, could put CEs together in groups based on edge-to-edge distance

    if look in {"f", "forward"}:
        warnings.warn("forward `look` considered experimental")

    css: list[geopandas.GeoDataFrame] = []
    for i in itimes:
        cs_i = contours_sets[i]
        cs_i["time"] = times[i]  # actual time
        cs_i["itime"] = itimes[i]  # time index (from 0)
        cs_i["dtime"] = dt[i]  # delta time
        n_i = len(cs_i)
        if i == 0:
            # IDs all new for first time step
            cs_i["mcs_id"] = range(n_i)
            next_id = n_i
        else:
            # Assign IDs using overlap threshold
            cs_im1 = css[i - 1]
            dt_im1_s = dt[i - 1].total_seconds()

            if look in {"b", "back"}:
                if overlap_norm is None:
                    overlap_norm = "a"
                ovs = overlap(
                    cs_i,
                    project(cs_im1, u=u_projection, dt=dt_im1_s),
                    norm=overlap_norm,
                )
                ids = []
                for j, d in ovs.items():
                    # For each CE at current time ("kid"/"child"),
                    # find previous CE of maximum overlap (single "parent")
                    k, frac = max(d.items(), key=lambda tup: tup[1], default=(None, 0))
                    if k is None or frac < overlap_threshold:
                        # No parent or not enough overlap => new ID
                        ids.append(next_id)
                        next_id += 1
                    else:
                        # Has parent; use their family ID
                        ids.append(cs_im1.loc[k].mcs_id)

                assert len(ids) == len(cs_i)

                if largest:
                    # For current CEs, make sure no MCS ID is shared (give to largest)
                    sz = cs_i["area_km2"].to_numpy(dtype=np.float64)
                    for mcs_id in set(ids):
                        inds_with_id = [j for j, id_ in enumerate(ids) if id_ == mcs_id]
                        if len(inds_with_id) == 1:
                            continue
                        logger.info(f"multiple CEs with same MCS ID: {inds_with_id}")
                        ind_largest, _ = max(
                            zip(inds_with_id, sz[inds_with_id]),
                            key=lambda tup: tup[1],
                        )
                        logger.info(
                            f"largest: {sz[ind_largest]} out of {sz[inds_with_id]}"
                        )
                        for j in inds_with_id:
                            if j == ind_largest:
                                continue
                            ids[j] = next_id
                            next_id += 1
                        assert ids[ind_largest] == mcs_id

            elif look in {"f", "forward"}:
                # Following `look='b'`,
                # `j`: iloc of CE at current time (i, `cs_i`)
                # `k`: iloc of CE at previous time (i-1, `cs_im1`)
                if overlap_norm is None:
                    overlap_norm = "b"
                ovs = overlap(
                    project(cs_im1, u=u_projection, dt=dt_im1_s),
                    cs_i,
                    norm=overlap_norm,
                )
                ids: list[int | None] = [None for _ in range(len(cs_i))]  # type: ignore[no-redef]
                sz_i = cs_i["area_km2"].to_numpy(dtype=np.float64)
                for k, d in ovs.items():
                    # For each CE at previous time ("parent"),
                    # look at CEs of current time that overlap ("kid"/"child")
                    mcs_id = cs_im1.loc[k].mcs_id
                    if not d:
                        continue

                    if largest and len(d) > 1:
                        # NOTE: 1-1 track not guaranteed since multiple parents could have same ID
                        # j, frac = max(d.items(), key=lambda tup: tup[1])  # max overlap
                        js = list(d.keys())
                        logger.info(f"{len(d)} possible children: {js}")
                        sz_ji = sz_i[js]
                        j, frac, _ = max(
                            zip(d.keys(), d.values(), sz_ji),
                            key=lambda tup: tup[2],
                        )  # max child area
                        assert sz_i[j] == sz_ji.max()
                        logger.info(f"keeping largest: {sz_i[j]} out of {sz_ji}")
                        d = {j: frac}

                    for j, frac in d.items():
                        if frac >= overlap_threshold:
                            if ids[j] is not None:
                                logger.info(
                                    f"warning: {j} already set to ID {ids[j]}, now {mcs_id}"
                                )
                                # TODO: support multiple parent
                            # Assign child the MCS ID of parent
                            ids[j] = mcs_id

                # For current CEs with no assigned parent, give new IDs
                for j, mcs_id in enumerate(ids):
                    if mcs_id is None:
                        ids[j] = next_id
                        next_id += 1

            else:
                raise ValueError("invalid `look`")

            cs_i["mcs_id"] = ids

        css.append(cs_i)

    # Combine into one frame
    cs = pd.concat(css)

    return cs.reset_index(drop=True)  # drop nested time, CE ind index


def calc_ellipse_eccen(p: shapely.geometry.polygon.Polygon):
    """Compute the (first) eccentricity of the least-squares best-fit ellipse
    to the coordinates of the polygon's exterior.
    """
    # TODO: Ellipse class with methods to convert to shapely Polygon/LinearRing and mpl Ellipse Patch

    # using skimage https://scikit-image.org/docs/dev/api/skimage.measure.html#skimage.measure.EllipseModel
    from skimage.measure import EllipseModel

    xy = np.asarray(p.exterior.coords)
    assert xy.shape[1] == 2

    m = EllipseModel()
    success = m.estimate(xy)

    if not success:
        warnings.warn(f"ellipse model failed for {p}")
        return np.nan

    _, _, xhw, yhw, _ = m.params
    # ^ xc, yc, a, b, theta; from the docs
    #   a with x, b with y (after subtracting the rotation), but they are half-widths
    #   theta is in radians

    rat = yhw / xhw if xhw > yhw else xhw / yhw

    return np.sqrt(1 - rat**2)


def _classify_one(cs: geopandas.GeoDataFrame) -> str:
    """Classify one CE family group."""
    # eps = sqrt(1 - (b^2/a^2)) -- ellipse "first eccentricity"
    #
    # Below from most to least strict:
    #
    # MCCs (organized)
    # - 219 K region >= 25k km2
    # - 235 K region >= 50k km2
    # - size durations have to be met for >= 6 hours
    # - eps <= 0.7
    #
    # CCCs (organized)
    # - 219 K region >= 25k km2
    # - size durations have to be met for >= 6 hours
    # - no shape criterion
    #
    # DLL (disorganized)
    # - >= 6 hour duration
    # - (no size or shape criterion)
    #
    # DSL (disorganized)
    # - < 6 hour duration
    #
    # Classification is for the "family" groups

    assert cs.mcs_id.unique().size == 1, "this is for a certain CE family group"

    # Sum areas over cloud elements
    time_groups = cs.groupby("time")
    area = time_groups[["area_km2", "area219_km2"]].apply(sum)

    # Get duration (time resolution of our CE data)
    dt = time_groups["dtime"].apply(_the_unique)
    dur_tot = dt.sum()

    # Compute area-duration criteria
    dur_219_25k = dt[area.area219_km2 >= 25_000].sum()
    dur_235_50k = dt[area.area_km2 >= 50_000].sum()
    six_hours = pd.Timedelta(hours=6)

    if dur_219_25k >= six_hours:  # organized
        # Compute ellipse eccentricity
        eps = time_groups[["geometry"]].apply(
            lambda g: calc_ellipse_eccen(g.dissolve().geometry.convex_hull.iloc[0])
        )
        dur_eps = dt[eps <= 0.7].sum()
        if dur_235_50k >= six_hours and dur_eps >= six_hours:
            class_ = "MCC"
        else:
            class_ = "CCC"

    else:  # disorganized
        if dur_tot >= six_hours:
            class_ = "DLL"
        else:
            class_ = "DSL"

    return class_


def classify(cs: geopandas.GeoDataFrame) -> geopandas.GeoDataFrame:
    """Classify the CE groups into MCS classes, adding a categorical ``'mcs_class'`` column
    to the input frame.
    """

    assert {"mcs_id", "time", "dtime"} < set(cs.columns), "needed by the classify algo"

    classes = cs.groupby("mcs_id").apply(_classify_one)
    cs["mcs_class"] = cs.mcs_id.map(classes).astype("category")

    return cs


def run(
    ds: xarray.DataArray,
    *,
    parallel: bool = True,
    u_projection: float = 0,
    ctt_threshold: float = 235,
    ctt_core_threshold: float = 219,
    **args,
) -> tuple[geopandas.GeoDataFrame, geopandas.GeoDataFrame, geopandas.GeoDataFrame]:
    r"""Run all TAMS steps, including precip assignment.

    .. important::
       `ds` must have ``'ctt'`` (cloud-top temperature) and ``'pr'`` (precip rate) variables.
       Dims should be ``'time'``, ``'lat'``, ``'lon'``.
       ``'lon'`` should be in -180 -- 180 format.


    Usage:

    >>> ce, mcs, mcs_summary = tams.run(ds)

    Parameters
    ----------
    ds
        Dataset containing 3-D cloud-top temperature and precipitation rate.
    parallel
        Whether to apply parallelization (where possible).
    u_projection
        *x*\-direction projection velocity to apply before computing overlaps.
    ctt_threshold
        Used to identify the edges of cloud elements.
    ctt_core_threshold
        Used to identify deep convective cloud regions within larger cloud areas.
        This is used to determine whether or not a system is eligible for being classified
        as an organized system.
        It helps target raining clouds.

    See Also
    --------
    :func:`tams.identify`
    :func:`tams.track`
    :func:`tams.classify`

    :doc:`/examples/tams-run`
    """
    import itertools

    import geopandas as gpd
    from shapely.errors import ShapelyDeprecationWarning
    from shapely.geometry import MultiPolygon

    assert "time" in ds.dims

    # TODO: timing and progress indicators, possibly with Rich

    def msg(s):
        """Print message and current time"""
        import datetime

        st = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        print(s, st)

    #
    # 1. Identify
    #

    msg("Starting `identify`")
    cs235, cs219 = identify(
        ds.ctt,
        parallel=parallel,
        ctt_threshold=ctt_threshold,
        ctt_core_threshold=ctt_core_threshold,
    )

    # If there are no tracked cloud elements then return none
    if sum([(cs235[i].empty) for i in range(len(cs235))]) == len(cs235):
        return [None, None]

    #
    # 2. Track
    #

    msg("Starting `track`")
    times = ds.time.values
    dt = pd.Timedelta(times[1] - times[0])  # TODO: equal spacing check here?
    ce = track(cs235, times, u_projection=u_projection)

    #
    # 3. Classify
    #

    msg("Starting `classify`")
    ce = classify(ce)

    latlon_coord_system = (ds.lat, ds.lon)

    return ce, latlon_coord_system

    # This assertion was on line 1116, but I moved it since precipitation rate isn't used until the mcs analysis
    # pure identification, tracking, and classification only depends on brightness temperature
    assert {"ctt", "pr"} <= set(ds.data_vars)
    # TODO : there is an error with transforming the xarray to a df in data_in_contours_sjoin, fix it later but for now just return the cloud elements
    #
    # 4. Stats (including precip)
    #

    msg("Starting statistics calculations")

    # Cleanup
    ce = ce.drop(columns=["inds219", "itime", "dtime"]).convert_dtypes()
    ce = ce.rename(columns={"geometry": "cs235"}).set_geometry("cs235")
    ce.cs219 = ce.cs219.set_crs("EPSG:4326")  # TODO: ensure set in `identify`

    msg("Starting CE aggregation (into MCS time series)")
    dfs_t = []
    ds_nt = []
    for mcs_id, mcs_ in ce.groupby("mcs_id"):
        # Time-varying
        time_group = mcs_.groupby("time")
        d = {}

        with warnings.catch_warnings():
            # ShapelyDeprecationWarning: __len__ for multi-part geometries is deprecated ...
            warnings.filterwarnings(
                "ignore",
                category=ShapelyDeprecationWarning,
                message="__len__ for multi-part geometries is deprecated",
            )
            d["cs235"] = gpd.GeoSeries(
                time_group.apply(lambda g: MultiPolygon(g.geometry.values))
            )
            d["cs219"] = gpd.GeoSeries(
                time_group.apply(
                    lambda g: MultiPolygon(
                        itertools.chain.from_iterable(mp.geoms for mp in g.cs219.values)
                    )
                )
            )

        d["nce"] = len(mcs_)
        d["area_km2"] = time_group.area_km2.sum()
        d["area219_km2"] = time_group.area219_km2.sum()
        # TODO: compare to re-computing area after (could be different if shift to dissolve)?

        df = pd.DataFrame(d).reset_index()  # time -> column
        df["mcs_id"] = mcs_id
        assert mcs_.mcs_class.unique().size == 1
        df["mcs_class"] = mcs_.mcs_class.values[0]

        # Summary stuff
        d2 = {}
        times = mcs_.time.unique()
        d2["first_time"] = times.min()
        d2["last_time"] = times.max()
        d2["duration"] = d2["last_time"] - d2["first_time"] + dt
        d2["mcs_id"] = mcs_id
        d2["mcs_class"] = mcs_.mcs_class.values[0]

        dfs_t.append(df)
        ds_nt.append(d2)

    # Initial MCS time-resolved
    mcs = (
        gpd.GeoDataFrame(pd.concat(dfs_t).reset_index(drop=True))
        .set_geometry("cs235", crs="EPSG:4326")
        .convert_dtypes()
    )
    mcs.cs219 = mcs.cs219.set_crs("EPSG:4326")
    mcs.mcs_class = mcs.mcs_class.astype("category")

    # Add CTT and PR data stats (time-resolved)
    msg("Starting gridded data aggregation")

    def _agg_one(ds_t, g):
        df1 = data_in_contours(ds_t.pr, g, merge=True)
        df2 = data_in_contours(
            ds_t.pr, g.set_geometry("cs219", drop=True), merge=False
        ).add_suffix("219")
        df3 = data_in_contours(
            ds_t.ctt, g.set_geometry("cs219", drop=True), merge=False
        ).add_suffix("219")
        df = (
            df1.join(df2)
            .join(df3)
            .drop(
                columns=[
                    "count_pr219",
                ]
            )
            .rename(columns={"count_pr": "npixel", "count_ctt219": "npixel219"})
        )
        return df

    if parallel:
        try:
            import joblib
        except ImportError as e:
            raise RuntimeError("joblib required") from e

        # TODO: Sometimes getting
        # > UserWarning: A worker stopped while some jobs were given to the executor.
        # > This can be caused by a too short worker timeout or by a memory leak.
        # Increasing `batch_size` reduces the number of these (e.g. to 1 with batch 10, 119 jobs, 11 workers).
        # Probably better to leave on auto to keep more general though.
        # Run time doesn't seem affected.
        print(f"ds is {ds}")
        for t, g in mcs.groupby("time"):
            dfs_app = _agg_one(ds.sel(time=t), g)
            print(f"the appended quantity on line 1274 is {dfs_app}")
        dfs = joblib.Parallel(n_jobs=-2, verbose=10, batch_size="auto")(
            joblib.delayed(_agg_one)(ds.sel(time=t).copy(deep=True), g.copy())
            for t, g in mcs.groupby("time")
        )
    else:
        dfs = [_agg_one(ds.sel(time=t), g) for t, g in mcs.groupby("time")]

    mcs = pd.concat(dfs)

    # Initial MCS summary
    mcs_summary = pd.DataFrame(ds_nt).reset_index(drop=True).convert_dtypes()
    mcs_summary.mcs_class = mcs_summary.mcs_class.astype("category")

    # Add some CTT and PR stats to summary dataset
    msg("Computing stats for MCS summary dataset")
    vns = [
        "mean_pr",
        "mean_pr219",
        "mean_ctt219",
        "std_ctt219",
        "area_km2",
        "area219_km2",
    ]
    mcs_summary = mcs_summary.join(
        mcs.groupby("mcs_id")[vns]
        .mean()
        .rename(columns={vn: f"mean_{vn}" for vn in vns})
    )

    # Add first and last points and distance to MCS summary dataset,
    # setting first point as the `geometry`

    def f(g):
        g.sort_values(by="time")  # should be already but just in case...
        cen = g.geometry.to_crs("EPSG:32663").centroid.to_crs("EPSG:4326")
        return gpd.GeoSeries(
            {"first_centroid": cen.iloc[0], "last_centroid": cen.iloc[-1]}
        )

    mcs_summary_points = gpd.GeoDataFrame(
        mcs.groupby("mcs_id").apply(f).astype("geometry")
    )
    # ^ Initially we have GeoDataFrame but the columns don't have dtype geometry
    # `.astype("geometry")` makes that conversion but we lose GeoDataFrame

    # `.set_crs()` only works on a dtype=geometry column in a GeoDataFrame
    mcs_summary_points.first_centroid = mcs_summary_points.first_centroid.set_crs(
        "EPSG:4326"
    )
    mcs_summary_points.last_centroid = mcs_summary_points.last_centroid.set_crs(
        "EPSG:4326"
    )
    assert (
        mcs_summary_points.first_centroid.crs
        == mcs_summary_points.last_centroid.crs
        == "EPSG:4326"
    )

    mcs_summary_points["distance_km"] = (
        mcs_summary_points.first_centroid.to_crs("EPSG:32663").distance(
            mcs_summary_points.last_centroid.to_crs("EPSG:32663")
        )
        / 10**3
    ).astype("Float64")

    mcs_summary = (
        gpd.GeoDataFrame(mcs_summary)
        .join(mcs_summary_points)
        .set_geometry("first_centroid")
    )

    # Final cleanup
    mcs = mcs.reset_index(drop=True)
    mcs_summary = mcs_summary.reset_index(drop=True)

    msg("Done")

    return ce, mcs, mcs_summary


# if __name__ == "__main__":
#     import cartopy.crs as ccrs
#     import matplotlib.pyplot as plt
#     import regionmask

#     from .data import load_example_ir, tb_from_ir

#     r = load_example_ir().isel(time=0)

#     tb = tb_from_ir(r, ch=9)

#     tran = ccrs.PlateCarree()
#     proj = ccrs.Mercator()
#     fig, ax = plt.subplots(subplot_kw=dict(projection=proj))

#     # tb.plot(x="lon", y="lat", cmap="gray_r", ax=ax)
#     cs = contours(tb, 235)
#     cs = sorted(cs, key=len, reverse=True)  # [:30]
#     for c in cs:
#         ax.plot(c[:, 0], c[:, 1], "g", transform=tran)

#     cs235 = _contours_to_gdf(cs)
#     cs219 = _contours_to_gdf(contours(tb, 219))

#     cs235, cs219 = _size_filter_contours(cs235, cs219)

#     # Trying regionmask
#     shapes = cs235[["geometry"]]
#     regions = regionmask.from_geopandas(shapes)
#     mask = regions.mask(tb)

#     regions.plot(ax=ax)

#     # tb.where(mask >= 0).plot.pcolormesh(ax=ax, transform=tran)  # takes long
#     tb.where(mask >= 0).plot.pcolormesh(size=4, aspect=2)

#     plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 10 16:28:49 2024

@author: thahn
"""

# =============================================================================
# Loads in and grids the NEXRAD Arhcival Level 2 Data using ARM-Pyart then converts into iris cubes and xarray Datasets for use in trackers
# =============================================================================


import glob
import multiprocessing
import os
import warnings
from functools import partial

import cftime
import iris
import iris.cube
import numpy as np
import pyart
import xarray as xr
from tqdm import tqdm


def gen_and_save_nexrad_grid(
    path_to_files: str,
    save_location: str,
    tracking_var: str,
    CONFIG: dict,
    parallel_processing: bool = False,
    max_cores: int | None = None,
) -> None:
    """


    Parameters
    ----------
    path_to_files : str
        Glob type path to the NEXRAD level 2 archive files.
    save_location : str
        Path to where the gridded NEXRAD files should be saved to, should be a directory and end with "/".
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity.
    CONFIG : dict
        User configuration file.
    parallel_processing : bool, optional
        Bool determinig whether to use parallel processing when gridding files. The default is False.
    max_cores : int, optional
        Number of cores to use if parallel_processing == True. The default is None.

    Raises
    ------
    Exception
        Exception if no files to grid or invalid tracking variables.

    Returns
    -------
    None

    """

    # Get all archive files and iterate over them
    files = np.sort(glob.glob(path_to_files))

    if len(files) == 0:
        raise Exception("!=====No Files Present to Grid=====!")
        return

    # Create grid folder if it does not already exist
    if not os.path.exists(save_location):
        os.makedirs(save_location)

    # If parallel processing is enabled, run that version and return
    if parallel_processing:
        gen_and_save_nexrad_grid_multi(
            files, save_location, tracking_var, CONFIG, max_cores
        )
        return

    # Extract just the filenames from the paths without the file extensions
    file_names = [os.path.basename(f).split(".")[0] for f in files]

    for idx, ff in tqdm(
        enumerate(files), desc="=====Gridding NEXRAD=====", total=len(files)
    ):
        if tracking_var.lower() == "dbz":
            # Create radar object including only field of interest
            radar = pyart.io.read_nexrad_archive(ff, include_fields="reflectivity")

            # Create radar grid using user-defined params
            radar_grid = pyart.map.grid_from_radars(
                radar, **CONFIG["nexrad"]["gridding"]
            )

            # Save radar grid to save_location as a netcdf file and delete radar and radar_grid objects to free memory
            pyart.io.write_grid(
                save_location + file_names[idx] + "_grid.nc",
                radar_grid,
                arm_alt_lat_lon_variables=True,
                write_point_x_y_z=True,
                write_point_lon_lat_alt=True,
            )

            del radar
            del radar_grid

        else:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )
            return

    return


def create_and_save_grid_single(file, save_location, tracking_var, CONFIG):
    """
    Helper Function for Parallel Processing
    """

    if tracking_var.lower() == "dbz":
        # Create radar object including only field of interest
        radar = pyart.io.read_nexrad_archive(file, include_fields="reflectivity")

        # Create radar grid using user-defined params
        radar_grid = pyart.map.grid_from_radars(radar, **CONFIG["nexrad"]["gridding"])

        # Save radar grid to save_location as a netcdf file and delete radar and radar_grid objects to free memory
        file_name = os.path.basename(file).split(".")[0]
        pyart.io.write_grid(
            save_location + file_name + "_grid.nc",
            radar_grid,
            arm_alt_lat_lon_variables=True,
            write_point_x_y_z=True,
            write_point_lon_lat_alt=True,
        )

        del radar
        del radar_grid

    else:
        raise Exception(
            f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
        )
        return


def gen_and_save_nexrad_grid_multi(
    files: list[str],
    save_location: str,
    tracking_var: str,
    CONFIG: dict,
    max_cores: int,
) -> None:
    """


    Parameters
    ----------
    files : list[str]
        List containing all paths to NEXRAD level 2 archive files.
    save_location : str
        Path to where the gridded NEXRAD files should be saved to.
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity.
    CONFIG : dict
        User configuration file.
    max_cores : int
        Number of cores to use for parallel processing

    Returns
    -------
    None

    """

    # Start a pool with max_cores and run the grid function
    with multiprocessing.Pool(max_cores) as multi_pool:
        with tqdm(total=len(files), desc="=====Parallel Gridding NEXRAD=====") as pbar:
            for _ in multi_pool.imap_unordered(
                partial(
                    create_and_save_grid_single,
                    save_location=save_location,
                    tracking_var=tracking_var,
                    CONFIG=CONFIG,
                ),
                files,
            ):
                pbar.update()

    return


def nexrad_load_netcdf_iris(
    path_to_files: str,
    file_type: str,
    tracking_var: str,
    CONFIG: dict,
    save_location: str | None = None,
) -> tuple[iris.cube.Cube, xr.DataArray]:
    """


    Parameters
    ----------
    path_to_files : str
        Glob path to input files, either archival or grided netcdf--i.e. "/data/usr/KVNX*_V06.ar2v".
    file_type : str
        ["ar2v", "nc"] type of input file--either archival or netcdf.
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity.
    CONFIG : dict
        User configuration file.
    save_location : str, optional
        Where to save gridded NEXRAD data to if file_type=="ar2v". The default is None.

    Returns
    -------
    cube : iris.cube.Cube
        Iris cube continaing gridded reflectivity data ready for tobac tracking.
    nexrad_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing gridded NEXRAD archival data.

    """

    # If data is archival, perform gridding
    if file_type.lower() == "ar2v":
        # Make sure save_location ends with a "/"
        if save_location[-1] != "/":
            save_location = save_location + "//"

        # Create grid
        gen_and_save_nexrad_grid(
            path_to_files,
            save_location,
            tracking_var,
            CONFIG,
            CONFIG["parallel_processing"],
            CONFIG["max_cores"],
        )

        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(save_location + "*")):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity.drop_vars(
                [
                    "origin_altitude",
                    "origin_latitude",
                    "origin_longitude",
                    "ProjectionCoordinateSystem",
                    "projection",
                ]
            )
            del radar_objects

            # Subset location and time of interest
            if "nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["nexrad"]
                    or "max_frame_index" in CONFIG["nexrad"]
                ):
                    min_frame = (
                        CONFIG["nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["nexrad"]:
                    mask_lon = (nexrad_xarray.lon >= CONFIG["nexrad"]["bounds"][0]) & (
                        nexrad_xarray.lon <= CONFIG["nexrad"]["bounds"][1]
                    )
                    mask_lat = (nexrad_xarray.lat >= CONFIG["nexrad"]["bounds"][2]) & (
                        nexrad_xarray.lat <= CONFIG["nexrad"]["bounds"][3]
                    )

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[3])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[3])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
            )

            # Create DT attribute
            dt_array = np.diff(nexrad_xarray.time.values)
            if len(np.unique(dt_array)) != 1:
                nexrad_xarray.attrs["DT"] = dt_array * 60  # min -> s
            else:
                nexrad_xarray.attrs["DT"] = dt_array[0] * 60  # min -> s

            # Adjust dimension names to be standards accepted by iris
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )

            # Add altitude dimension to xarray but not to cube
            nexrad_cube = nexrad_xarray.to_iris()
            nexrad_xarray = nexrad_xarray.assign_coords(
                altitude=("z", nexrad_xarray.z.values)
            )

            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs({"standard_name": ""})
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return (nexrad_cube, nexrad_xarray)

        else:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )
            return

    # If data already grided, just return concated netcdf dataset
    elif file_type.lower() == "nc":
        # Convert to iris cube and return
        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(path_to_files)):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity.drop_vars(
                [
                    "origin_altitude",
                    "origin_latitude",
                    "origin_longitude",
                    "ProjectionCoordinateSystem",
                    "projection",
                ]
            )
            del radar_objects

            # Subset location and time of interest
            if "nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["nexrad"]
                    or "max_frame_index" in CONFIG["nexrad"]
                ):
                    min_frame = (
                        CONFIG["nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["nexrad"]:
                    mask_lon = (nexrad_xarray.lon >= CONFIG["nexrad"]["bounds"][0]) & (
                        nexrad_xarray.lon <= CONFIG["nexrad"]["bounds"][1]
                    )
                    mask_lat = (nexrad_xarray.lat >= CONFIG["nexrad"]["bounds"][2]) & (
                        nexrad_xarray.lat <= CONFIG["nexrad"]["bounds"][3]
                    )

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[3])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[3])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
            )

            # Create DT attribute
            dt_array = np.diff(nexrad_xarray.time.values)
            if len(np.unique(dt_array)) != 1:
                nexrad_xarray.attrs["DT"] = dt_array * 60  # min -> s
            else:
                nexrad_xarray.attrs["DT"] = dt_array[0] * 60  # min -> s

            # Adjust dimension names to be standards accepted by iris
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )

            # Add altitude dimension to xarray but not to cube
            nexrad_cube = nexrad_xarray.to_iris()  # .remove_coord("altitude")
            nexrad_xarray = nexrad_xarray.assign_coords(
                altitude=("z", nexrad_xarray.z.values)
            )

            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs({"standard_name": ""})
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return (nexrad_cube, nexrad_xarray)

        else:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )
            return

    else:
        raise Exception(
            f"!=====Invalid File Type. You Entered: {file_type.lower()}=====!"
        )
        return


def nexrad_load_netcdf(
    path_to_files: str,
    file_type: str,
    tracking_var: str,
    CONFIG: dict,
    save_location: str | None = None,
) -> xr.DataArray:
    """


    Parameters
    ----------
    path_to_files : str
        Glob path to input files, either archival or grided netcdf--i.e. "/data/usr/KVNX*_V06.ar2v".
    file_type : str
        ["ar2v", "nc"] type of input file--either archival or netcdf.
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity.
    CONFIG : dict
        User configuration file.
    save_location : str, optional
        Where to save gridded NEXRAD data to if file_type=="ar2v". The default is None.

    Returns
    -------
    nexrad_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing gridded NEXRAD archival data.

    """

    # If data is archival, perform gridding
    if file_type.lower() == "ar2v":
        # Make sure save_location ends with a "/"
        if save_location[-1] != "/":
            save_location = save_location + "//"

        # Create grid
        gen_and_save_nexrad_grid(
            path_to_files,
            save_location,
            tracking_var,
            CONFIG,
            CONFIG["parallel_processing"],
            CONFIG["max_cores"],
        )

        # Open them as netcdf file and return
        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(save_location + "*")):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity.drop_vars(
                [
                    "origin_altitude",
                    "origin_latitude",
                    "origin_longitude",
                    "ProjectionCoordinateSystem",
                    "projection",
                ]
            )
            del radar_objects

            # Subset location and time of interest
            if "nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["nexrad"]
                    or "max_frame_index" in CONFIG["nexrad"]
                ):
                    min_frame = (
                        CONFIG["nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["nexrad"]:
                    mask_lon = (nexrad_xarray.lon >= CONFIG["nexrad"]["bounds"][0]) & (
                        nexrad_xarray.lon <= CONFIG["nexrad"]["bounds"][1]
                    )
                    mask_lat = (nexrad_xarray.lat >= CONFIG["nexrad"]["bounds"][2]) & (
                        nexrad_xarray.lat <= CONFIG["nexrad"]["bounds"][3]
                    )

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[3])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[3])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
                altitude=("z", nexrad_xarray.z.values),
            )

            # Adjust dimension names to be standards accepted by iris
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return nexrad_xarray

        else:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )
            return

    # If data already grided, just return concated netcdf dataset
    elif file_type.lower() == "nc":
        if tracking_var.lower() == "dbz":
            # Read radar objects in and concat into one xarray datarray
            # This is a stupid hacky fix because pyart is dumb
            radar_objects = []
            for file in np.sort(glob.glob(path_to_files)):
                # Ignore import warnings
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    radar_objects.append(pyart.io.read_grid(file).to_xarray())

            nexrad_xarray = xr.concat(radar_objects, dim="time").reflectivity.drop_vars(
                [
                    "origin_altitude",
                    "origin_latitude",
                    "origin_longitude",
                    "ProjectionCoordinateSystem",
                    "projection",
                ]
            )
            del radar_objects

            # Subset location and time of interest
            if "nexrad" in CONFIG:
                # Subset time based on user inputs
                if (
                    "min_frame_index" in CONFIG["nexrad"]
                    or "max_frame_index" in CONFIG["nexrad"]
                ):
                    min_frame = (
                        CONFIG["nexrad"]["min_frame_index"]
                        if "min_frame_index" in CONFIG["nexrad"]
                        else 0
                    )
                    max_frame = (
                        CONFIG["nexrad"]["max_frame_index"] + 1
                        if "max_frame_index" in CONFIG["nexrad"]
                        else nexrad_xarray.dims["time"]
                    )

                    nexrad_xarray = nexrad_xarray.isel(
                        time=np.arange(
                            min_frame,
                            max_frame,
                        ),
                        drop=True,
                    )

                if "bounds" in CONFIG["nexrad"]:
                    mask_lon = (nexrad_xarray.lon >= CONFIG["nexrad"]["bounds"][0]) & (
                        nexrad_xarray.lon <= CONFIG["nexrad"]["bounds"][1]
                    )
                    mask_lat = (nexrad_xarray.lat >= CONFIG["nexrad"]["bounds"][2]) & (
                        nexrad_xarray.lat <= CONFIG["nexrad"]["bounds"][3]
                    )

                    nexrad_xarray = nexrad_xarray.where(mask_lon & mask_lat, drop=True)

            else:
                raise Exception("""!=====CONFIG Missing "nexrad" Field=====!""")

            # Replace time dimension with minutes since first time and add other x y z coords
            first_time = nexrad_xarray.time.values[0]
            nexrad_xarray = nexrad_xarray.assign_coords(
                time=cftime.date2num(
                    nexrad_xarray.time.values, f"minutes since {first_time}"
                ),
                south_north=("y", np.arange(nexrad_xarray.shape[2])),
                west_east=("x", np.arange(nexrad_xarray.shape[3])),
                projection_x_coordinate=("x", nexrad_xarray.x.values),
                projection_y_coordinate=("y", nexrad_xarray.y.values),
                x=("x", np.arange(nexrad_xarray.shape[3])),
                y=("y", np.arange(nexrad_xarray.shape[2])),
                model_level_number=("z", np.arange(nexrad_xarray.shape[1])),
                altitude=("z", nexrad_xarray.z.values),
            )

            # Adjust dimension names to be standards
            nexrad_xarray["time"] = nexrad_xarray.time.assign_attrs(
                {
                    "standard_name": "time",
                    "long_name": f"minutes since {first_time}",
                    "units": f"minutes since {first_time}",
                }
            )
            nexrad_xarray["z"] = nexrad_xarray.z.assign_attrs(
                {"standard_name": "", "units": "m"}
            )
            nexrad_xarray["lat"] = nexrad_xarray.lat.assign_attrs(
                {"standard_name": "latitude"}
            )
            nexrad_xarray["lon"] = nexrad_xarray.lon.assign_attrs(
                {"standard_name": "longitude"}
            )
            nexrad_xarray["projection_x_coordinate"] = (
                nexrad_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["projection_y_coordinate"] = (
                nexrad_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
            )
            nexrad_xarray["altitude"] = nexrad_xarray.altitude.assign_attrs(
                {"standard_name": "altitude", "units": "m"}
            )

            return nexrad_xarray

        else:
            raise Exception(
                f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
            )
            return

    else:
        raise Exception(
            f"!=====Invalid File Type. You Entered: {file_type.lower()}=====!"
        )
        return



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions for use outside of CoCoMET. If the user wishes to implement custom functionality based on the CoCoMET methodologies, these functions help facilitate that.
"""

from CoCoMET.user_utils.bulk_statistics_funcs import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jul 18 12:19:18 2024

@author: thahn
"""

#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 10 16:28:49 2024

@author: thahn
"""

# =============================================================================
# Loads in pre-gridded radar data which follows the radar standardization set out in the CoCoMET-US Section S1.1.
# =============================================================================

import iris.cube
import numpy as np
import xarray as xr


def standard_radar_load_netcdf_iris(
    path_to_files: str, tracking_var: str, CONFIG: dict
) -> tuple[iris.cube.Cube, xr.DataArray]:
    """


    Parameters
    ----------
    path_to_files : str
        Glob path to gridded input netcdf files--i.e. "/data/usr/KVNX*_V06.nc".
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity..
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if CONFIG missing standard_radar field or invalid tracking variable entered.

    Returns
    -------
    radar_cube : iris.cube.Cube
        Iris cube continaing gridded reflectivity data ready for tobac tracking.
    radar_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing gridded reflectivity data.

    """

    # Convert to iris cube and return
    if tracking_var.lower() == "dbz":
        # Open combined netcdf radar dataarray
        radar_xarray = xr.open_mfdataset(
            path_to_files,  # concat_dim="time", combine="nested"
        ).reflectivity

        # Subset location of interest
        if "standard_radar" in CONFIG:
            # Subset time based on user inputs
            if (
                "min_frame_index" in CONFIG["standard_radar"]
                or "max_frame_index" in CONFIG["standard_radar"]
            ):
                min_frame = (
                    CONFIG["standard_radar"]["min_frame_index"]
                    if "min_frame_index" in CONFIG["standard_radar"]
                    else 0
                )
                max_frame = (
                    CONFIG["standard_radar"]["max_frame_index"] + 1
                    if "max_frame_index" in CONFIG["standard_radar"]
                    else radar_xarray.dims["time"]
                )

                radar_xarray = radar_xarray.isel(
                    time=np.arange(
                        min_frame,
                        max_frame,
                    ),
                    drop=True,
                )

            if "bounds" in CONFIG["standard_radar"]:
                mask_lon = (
                    radar_xarray.lon >= CONFIG["standard_radar"]["bounds"][0]
                ) & (radar_xarray.lon <= CONFIG["standard_radar"]["bounds"][1])
                mask_lat = (
                    radar_xarray.lat >= CONFIG["standard_radar"]["bounds"][2]
                ) & (radar_xarray.lat <= CONFIG["standard_radar"]["bounds"][3])

                try:
                    radar_xarray = radar_xarray.where(mask_lon & mask_lat, drop=True)
                except:
                    radar_xarray = radar_xarray.where(
                        np.logical_and(mask_lon, mask_lat).compute(), drop=True
                    )

        else:
            raise Exception("""!=====CONFIG Missing "standard_radar" Field=====!""")

        # Create DT attribute
        dt_array = (
            np.diff(radar_xarray.time.values).astype("timedelta64[s]").astype(int)
        )
        if len(np.unique(dt_array)) != 1:
            radar_xarray.attrs["DT"] = dt_array
        else:
            radar_xarray.attrs["DT"] = dt_array[0]

        first_time = radar_xarray.time.values[0]

        radar_xarray = radar_xarray.assign_coords(
            time=(
                "time",
                (radar_xarray.time.values - first_time)
                .astype("timedelta64[m]")
                .astype(float),
            )
        )
        radar_xarray["time"] = radar_xarray.time.assign_attrs(
            {
                "standard_name": "time",
                "long_name": f"minutes since {first_time}",
                "units": f"minutes since {first_time}",
            }
        )

        # if not("projection_x_coordinate" in radar_xarray.coords) and not("projection_y_coordinate" in radar_xarray.coords):

        radar_xarray = radar_xarray.assign_coords(
            projection_x_coordinate=("x", radar_xarray.proj_x.values),
            projection_y_coordinate=("y", radar_xarray.proj_y.values),
        )

        radar_xarray["projection_x_coordinate"] = (
            radar_xarray.projection_x_coordinate.assign_attrs({"units": "m"})
        )
        radar_xarray["projection_y_coordinate"] = (
            radar_xarray.projection_y_coordinate.assign_attrs({"units": "m"})
        )

        # Drop altitude coordinate temporarily when making iris cube
        drop_variables = ["altitude", "proj_x", "proj_y"]
        for dv in drop_variables:
            if dv in radar_xarray.coords:
                radar_xarray = radar_xarray.drop_vars(dv)
        radar_cube = radar_xarray.to_iris()

        radar_xarray = radar_xarray.assign_coords(altitude=("z", radar_xarray.z.values))
        radar_xarray["z"] = radar_xarray.z.assign_attrs({"standard_name": ""})
        radar_xarray["altitude"] = radar_xarray.altitude.assign_attrs(
            {"standard_name": "altitude", "units": "m"}
        )

        return (radar_cube, radar_xarray)

    else:
        raise Exception(
            f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
        )


def standard_radar_load_netcdf(
    path_to_files: str, tracking_var: str, CONFIG: dict
) -> xr.DataArray:
    """


    Parameters
    ----------
    path_to_files : str
        Glob path to gridded input netcdf files--i.e. "/data/usr/KVNX*_V06.nc".
    tracking_var : str
        ["dbz"], variable which is going to be used for tracking--reflectivity..
    CONFIG : dict
        User configuration file.

    Raises
    ------
    Exception
        Exception if CONFIG missing standard_radar field or invalid tracking variable entered.

    Returns
    -------
    radar_xarray : xarray.core.dataarray.DataArray
        Xarray DataArray containing gridded reflectivity data.

    """

    # Convert to iris cube and return
    if tracking_var.lower() == "dbz":
        # Open combined netcdf radar dataarray
        radar_xarray = xr.open_mfdataset(
            path_to_files, concat_dim="time", combine="nested"
        ).reflectivity

        # Subset location of interest
        if "standard_radar" in CONFIG:
            # Subset time based on user inputs
            if (
                "min_frame_index" in CONFIG["standard_radar"]
                or "max_frame_index" in CONFIG["standard_radar"]
            ):
                min_frame = (
                    CONFIG["standard_radar"]["min_frame_index"]
                    if "min_frame_index" in CONFIG["standard_radar"]
                    else 0
                )
                max_frame = (
                    CONFIG["standard_radar"]["max_frame_index"] + 1
                    if "max_frame_index" in CONFIG["standard_radar"]
                    else radar_xarray.dims["time"]
                )

                radar_xarray = radar_xarray.isel(
                    time=np.arange(
                        min_frame,
                        max_frame,
                    ),
                    drop=True,
                )

            if "bounds" in CONFIG["standard_radar"]:
                mask_lon = (
                    radar_xarray.lon >= CONFIG["standard_radar"]["bounds"][0]
                ) & (radar_xarray.lon <= CONFIG["standard_radar"]["bounds"][1])
                mask_lat = (
                    radar_xarray.lat >= CONFIG["standard_radar"]["bounds"][2]
                ) & (radar_xarray.lat <= CONFIG["standard_radar"]["bounds"][3])

                radar_xarray = radar_xarray.where(mask_lon & mask_lat, drop=True)

        else:
            raise Exception("""!=====CONFIG Missing "standard_radar" Field=====!""")

        first_time = radar_xarray.time.values[0]
        radar_xarray = radar_xarray.assign_coords(
            time=(
                "time",
                (radar_xarray.time.values - first_time)
                .astype("timedelta64[m]")
                .astype(float),
            )
        )
        radar_xarray["time"] = radar_xarray.time.assign_attrs(
            {
                "standard_name": "time",
                "long_name": f"minutes since {first_time}",
                "units": f"minutes since {first_time}",
            }
        )

        radar_xarray["z"] = radar_xarray.z.assign_attrs({"standard_name": ""})
        radar_xarray["altitude"] = radar_xarray.altitude.assign_attrs(
            {"standard_name": "altitude", "units": "m"}
        )

        # Return subseted radar xarray
        return radar_xarray

    else:
        raise Exception(
            f"!=====Invalid Tracking Variable. You Entered: {tracking_var.lower()}=====!"
        )



import glob
from datetime import datetime

import numpy as np
import xarray as xr
from tqdm import tqdm


# TODO: Remove this and replace with RAMS-MAT package
def configure_rams(dataset0, path_to_header, CONFIG=None, configure_variables=[]):
    """
    Serves to set up proper attributes, coordinates, and dimensions so that the RAMS data set is compatible with CoCoMET
    Inputs:
        dataset: an xarray of RAMS data

    Outputs:
        configured_dataset: an xarray of RAMS data which has the proper attributes, coordinates, and dimensions
    """

    # avoid errors with CONFIG file
    if CONFIG is None:
        CONFIG = {"verbose": False, "rams": {}}

    # since grid 3 of RAMS dataset is large, create a list of only the necessary variables and a new dataset to reduce required computational power
    ListOfNeededVariables = [
        "GLAT",
        "GLON",
        "TOPT",  # lat, lon, and topographical height
        "PI",
        "THETA",  # pressure and temperature var
        # "RV",
        # "RCP",
        # "RDP",  # mixing ratio (dbz calculation)
        # "RRP",
        # "RPP",
        # "RSP",
        # "RAP",
        # "RGP",
        # "RHP",
        # "CCP",
        # "CDP",
        # "CRP",  # concentration numbers (dbz calculation)
        # "CPP",
        # "CSP",
        # "CAP",
        # "CGP",
        # "CHP",
        # "PCPVR",
        # "PCPVS",
        # "PCPVA",  # 3D precipitation rates
        # "PCPVG",
        # "PCPVH",
        # "PCPVD",
        # "LWUP",  # longwave radiation (brightness temp calculation)
        # "DN0",  # dry air density (experimental for precipitation rate calculation)
    ]

    # Add in the tracking variable(s)
    ListOfNeededVariables.extend(configure_variables)

    dataset = xr.Dataset({})
    if CONFIG["verbose"]:
        for var in tqdm(
            ListOfNeededVariables,
            desc="=====Creating Truncated RAMS Dataset=====",
            total=len(ListOfNeededVariables),
        ):

            try:
                dims = dataset0[var].dims
                if len(dims) == 3:
                    dataset[var] = ([*dims], dataset0[var].data[:, 1:, 1:])
                    dataset[var].chunk(
                        dataset0[var][:, 1:, 1:].chunksizes
                    )  # rechunk the variables
                elif len(dims) == 4:
                    dataset[var] = ([*dims], dataset0[var].data[:, 1:, 1:, 1:])
                    dataset[var].chunk(
                        dataset0[var][:, 1:, 1:, 1:].chunksizes
                    )  # rechunk the variables

            except:
                continue

    else:
        for var in ListOfNeededVariables:

            try:
                dims = dataset0[var].dims
                if len(dims) == 3:
                    dataset[var] = ([*dims], dataset0[var].data[:, 1:, 1:])
                    dataset[var].chunk(
                        dataset0[var][:, 1:, 1:].chunksizes
                    )  # rechunk the variables
                elif len(dims) == 4:
                    dataset[var] = ([*dims], dataset0[var].data[:, 1:, 1:, 1:])
                    dataset[var].chunk(
                        dataset0[var][:, 1:, 1:, 1:].chunksizes
                    )  # rechunk the variables]

            except:
                continue

    # destagger winds
    pbar = tqdm(
        np.arange(6),
        desc="=====Destaggering Winds and Adding Coordinates/Attributes=====",
        total=6,
    )
    windDims = dataset0["WC"].dims
    WC = 0.5 * (dataset0["WC"][:, 1:, 1:, 1:] + dataset0["WC"][:, :-1, 1:, 1:])
    dataset["WC"] = ([*windDims], WC.data)
    dataset["WC"].chunk(dataset0["WC"][:, 1:, 1:, 1:].chunksizes)
    pbar.update()
    pbar.refresh()
    VC = 0.5 * (dataset0["VC"][:, 1:, 1:, 1:] + dataset0["VC"][:, :-1, 1:, 1:])
    dataset["VC"] = ([*windDims], VC.data)
    dataset["VC"].chunk(dataset0["VC"][:, 1:, 1:, 1:].chunksizes)
    pbar.update()
    pbar.refresh()
    UC = 0.5 * (dataset0["UC"][:, 1:, 1:, 1:] + dataset0["UC"][:, :-1, 1:, 1:])
    dataset["UC"] = ([*windDims], UC.data)
    dataset["UC"].chunk(dataset0["UC"][:, 1:, 1:, 1:].chunksizes)
    pbar.update()
    pbar.refresh()

    # rename phony dimensions to dimensions that make sense
    dataset = dataset.rename(
        {
            "phony_dim_0": "south_north",
            "phony_dim_1": "west_east",
            "phony_dim_2": "bottom_top",
        }
    )  # , 'phony_dim_3' : 'np', 'phony_dim_4' : 'nzs', 'phony_dim_5' : 'local_topography_height'})
    pbar.update()
    pbar.refresh()

    # define the coordinates
    dataset = dataset.assign_coords(
        GLAT=(["Time", "south_north", "west_east"], dataset["GLAT"].data)
    )
    dataset = dataset.assign_coords(
        GLON=(["Time", "south_north", "west_east"], dataset["GLON"].data)
    )
    pbar.update()
    pbar.refresh()

    # Since RAMS does not have any attributes, add them. These are necessary for the iris cube generation
    dataset.attrs["CEN_LAT"] = np.median(dataset["GLAT"].values)
    dataset.attrs["CEN_LON"] = np.median(dataset["GLON"].values)
    dataset.attrs["WEST-EAST_PATCH_END_UNSTAG"] = len(dataset["west_east"])
    dataset.attrs["SOUTH-NORTH_PATCH_END_UNSTAG"] = len(dataset["south_north"])
    dataset.attrs["BOTTOM-TOP_PATCH_END_UNSTAG"] = len(dataset["bottom_top"])
    dataset.attrs["BOTTOM-TOP_PATCH_START_UNSTAG"] = 0
    dataset.attrs["SOUTH-NORTH_PATCH_START_UNSTAG"] = 0
    dataset.attrs["WEST-EAST_PATCH_START_UNSTAG"] = 0
    dataset.attrs["MAP_PROJ_CHAR"] = "Polar-Stereographic"
    pbar.update()
    pbar.refresh()

    # read in header file metadata
    # use two header files next to each other to find the time difference
    # read in grid spacing constants and altitudes
    listing = glob.glob(path_to_header)
    firstHeaderStr = listing[0]
    if len(listing) > 1:
        times = np.zeros(len(listing))
        datetime_times = []
        for ind, i in tqdm(
            enumerate(listing),
            desc="=====Loading RAMS Header File Data=====",
            total=len(listing),
        ):
            f = np.loadtxt(i, skiprows=findWhereColsChange(i), dtype=str)
            time_now = float(f[np.where(f == "__time")[0] + 2][0])
            times[ind] = time_now

            # use the last loaded file for the initial date
            year = int(f[np.where(f == "__iyear1")[0] + 2][0])
            month = int(f[np.where(f == "__imonth1")[0] + 2][0])
            date = int(f[np.where(f == "__idate1")[0] + 2][0])
            itime = f[np.where(f == "__itime1")[0] + 2][0]

            hour = int(np.floor(time_now / 3600))
            time_now -= hour * 3600
            minute = int(np.floor(time_now / 60))
            time_now -= minute * 60
            second = int(time_now)
            date_str_unformatted = f"{year}-{month}-{date} {int(itime[:2]) + hour}:{int(itime[2:]) + minute}:{second}"
            datetime_date = datetime.strptime(date_str_unformatted, "%Y-%m-%d %H:%M:%S")
            datetime_times.append(datetime_date)

            # declare the units from the initial time
            if ind == 0:
                date_str_formatted = datetime.strftime(
                    datetime_date, "%Y-%m-%d %H:%M:%S"
                )
                dataset.attrs["date"] = f"minutes since {date_str_formatted}"

        DT = times[1:] - times[:-1]
        if sum(~(DT == DT[0])) != 0:
            raise Exception("Header file time stamps are non-continuous")
        dataset.attrs["DT"] = DT[0]

    # Now that the times have been parsed, write the times coordinate
    dataset = dataset.assign_coords(Times=(["Time"], datetime_times))

    # use one header file to define the linear projection constants
    headerfile = np.loadtxt(
        firstHeaderStr, dtype=str, skiprows=findWhereColsChange(firstHeaderStr)
    )
    DX = float(headerfile[np.where(headerfile == "__xmn03")[0][0] + 3]) - float(
        headerfile[np.where(headerfile == "__xmn03")[0][0] + 2]
    )
    DY = float(headerfile[np.where(headerfile == "__ymn03")[0][0] + 3]) - float(
        headerfile[np.where(headerfile == "__ymn03")[0][0] + 2]
    )
    dataset.attrs["DY"] = DY
    dataset.attrs["DX"] = DX

    # calculate projection coordinates using linear projection constants
    proj_y_values = dataset.DY * (np.arange(0, dataset.south_north.shape[0]) + 0.5)
    proj_x_values = dataset.DX * (np.arange(0, dataset.west_east.shape[0]) + 0.5)

    dataset["PROJY"] = ("south_north", proj_y_values)
    dataset["PROJX"] = ("west_east", proj_x_values)

    # read in gamma shape parameters, used for reflectivity calculation
    IDgnu = np.where(headerfile == "__gnu")[0][0]
    numOfGnuPoints = int(headerfile[IDgnu + 1])
    startgnu = IDgnu + 2
    stopgnu = IDgnu + numOfGnuPoints + 2
    dataset.attrs["gnu"] = np.array(headerfile[startgnu:stopgnu], dtype=float)

    # calculate the altitude and geopotential
    unstagid = f"__zmn03"
    IDunstag = np.where(headerfile == unstagid)[0][0]

    numOfunstagPoints = int(headerfile[IDunstag + 1])
    startunstag = IDunstag + 2
    stopunstag = IDunstag + numOfunstagPoints + 2
    __zmn03 = np.array(headerfile[startunstag:stopunstag], dtype=float)[
        1:
    ]  # the first data point is unphysical

    # also add altitudes and geopotential to rams_xarray
    dataset["altitudes"] = (["bottom_top"], __zmn03)

    # if you are using moaap, calculate the geopotential
    if "moaap" in CONFIG["rams"]:
        TOPT = dataset["TOPT"].values
        geopt = np.zeros_like(dataset["THETA"])
        for t in tqdm(
            range(TOPT.shape[0]),
            desc="=====Calculating Geopotential for MOAAP Tracking=====",
            total=TOPT.shape[0],
        ):
            alt = np.array(
                np.expand_dims(dataset["altitudes"].values, (1, 2)), dtype=float
            )
            geopt[t, :, :, :] = (
                np.array(np.expand_dims(TOPT, 1)[t, :, :, :], dtype=float) + alt
            ) * 9.81
        geopt = xr.DataArray(
            geopt, dims=["Time", "bottom_top", "south_north", "west_east"]
        )
        geopt.chunk(dataset["THETA"].chunksizes)
        dataset["geopt"] = (
            ["Time", "bottom_top", "south_north", "west_east"],
            geopt.data,
        )

    return dataset


def findWhereColsChange(file):
    """
    Find where in a txt file the number of columns change for the first time,
    then return that index to be used as the "skiprows" argument of np.loadtxt

    Inputs:
        file: a txt file with varying column numbers

    Returns:
        displacer: an integer index of the first row in the txt file where the number of columns change
    """
    with open(file) as f:
        displacer = 1
        for line in f.readlines()[1:]:
            displacer += 1
            if len(line.split()) == 1:
                break
    return displacer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jun 24 12:34:22 2024

@author: thahn
"""

# =============================================================================
# Modifies tracking output into the CoCoMET-US
# =============================================================================

from copy import deepcopy
from datetime import datetime

import geopandas as gpd
import numpy as np
import pandas as pd
import xarray as xr
from matplotlib.path import Path
from scipy.interpolate import interpn
from scipy.ndimage import center_of_mass
from tqdm import tqdm


def feature_id_to_US(
    features: gpd.GeoDataFrame, tracker: str
) -> gpd.GeoDataFrame | None:
    """


    Parameters
    ----------
    features : geopandas.geodataframe.GeoDataFrame
        The output from the feature detection step of a given tracker.
    tracker : str
        ["tobac"] The tracker used to generate the features.

    Raises
    ------
    Exception
        Exception if invalid tracker entered.

    Returns
    -------
    US_features : geopandas.geodataframe.GeoDataFrame
        A geodataframe matching the format of the CoCoMET-US feature detection specification.

    """

    if features is None:
        return None

    if tracker.lower() == "tobac":
        # Extract values from features
        frames = features.frame.values
        times = np.array(
            [datetime.fromisoformat(f.isoformat()) for f in features.time.values]
        )
        feature_ids = features.feature.values - 1
        north_souths = features.south_north.values
        east_wests = features.west_east.values
        latitudes = features.latitude.values
        longitudes = features.longitude.values
        projection_x = features.projection_x_coordinate.values
        projection_y = features.projection_y_coordinate.values
        geometries = features.geometry.values

        # Include 3D coordinates if present. If not, set all alt values as NaN
        if "altitude" in features and "vdim" in features:
            altitudes = features.altitude.values / 1000
            up_downs = features.vdim.values

        else:
            altitudes = np.repeat(np.nan, features.shape[0])
            up_downs = np.repeat(np.nan, features.shape[0])

        # Create GeoDataFrame according to US specification
        US_features = gpd.GeoDataFrame(
            data={
                "frame": frames,
                "time": times,
                "feature_id": feature_ids,
                "south_north": north_souths,
                "west_east": east_wests,
                "up_down": up_downs,
                "latitude": latitudes,
                "longitude": longitudes,
                "projection_x": projection_x,
                "projection_y": projection_y,
                "altitude": altitudes,
                "geometry": geometries,
            }
        )

        return US_features

    raise Exception(f"!=====Invalid Tracker, You Entered: {tracker.lower()}=====!")


def linking_to_US(tracks: gpd.GeoDataFrame, tracker: str) -> gpd.GeoDataFrame | None:
    """


    Parameters
    ----------
    tracks : geopandas.geodataframe.GeoDataFrame
        The output from the linking/tracking step of a given tracker.
    tracker : str
        ["tobac"] The tracker used to generate the links.

    Raises
    ------
    Exception
        Exception if invalid tracker entered.

    Returns
    -------
    US_tracks : geopandas.geodataframe.GeoDataFrame
        A geodataframe matching the format of the CoCoMET-US linking specification.

    """

    # TODO: figure out how to remove cells with -1 cell id

    if tracks is None:
        return None

    if tracker.lower() == "tobac":
        # Extract values from features
        frames = tracks.frame.values
        times = np.array(
            [datetime.fromisoformat(f.isoformat()) for f in tracks.time.values]
        )
        lifetimes = tracks.time_cell.values
        feature_ids = tracks.feature.values - 1
        cell_ids = tracks.cell.values - 1
        # Correct any -2 values, created as a result of shifting, back to -1
        cell_ids[cell_ids == -2] = -1
        north_souths = tracks.south_north.values
        east_wests = tracks.west_east.values
        latitudes = tracks.latitude.values
        longitudes = tracks.longitude.values
        projection_x = tracks.projection_x_coordinate.values
        projection_y = tracks.projection_y_coordinate.values
        geometries = tracks.geometry.values

        # Include 3D coordinates if present. If not, set all alt values as NaN
        if "altitude" in tracks and "vdim" in tracks:
            altitudes = tracks.altitude.values / 1000
            up_downs = tracks.vdim.values

        else:
            altitudes = np.repeat(np.nan, tracks.shape[0])
            up_downs = np.repeat(np.nan, tracks.shape[0])

        lifetime_percents = []

        # Loop over rows
        for row in tqdm(
            tracks.itertuples(index=True),
            desc="=====Performing tobac Linking to US====",
            total=tracks.shape[0],
        ):
            cell_id = row.cell
            cell_max_life = tracks[tracks["cell"] == cell_id]["time_cell"].values.max()
            # cell_max_life = tracks.query("cell==@row[1].cell").time_cell.values.max()
            # If only tracked one time, add -1 to lifetime_percent
            if cell_max_life == 0:
                lifetime_percents.append(-1)
            else:
                lifetime_percents.append(row.time_cell / cell_max_life)

        # Create GeoDataFrame according to US specification
        US_tracks = gpd.GeoDataFrame(
            data={
                "frame": frames,
                "time": times,
                "lifetime": lifetimes,
                "lifetime_percent": lifetime_percents,
                "feature_id": feature_ids,
                "cell_id": cell_ids,
                "south_north": north_souths,
                "west_east": east_wests,
                "up_down": up_downs,
                "latitude": latitudes,
                "longitude": longitudes,
                "projection_x": projection_x,
                "projection_y": projection_y,
                "altitude": altitudes,
                "geometry": geometries,
            }
        )

        return US_tracks

    raise Exception(f"!=====Invalid Tracker, You Entered: {tracker.lower()}=====!")


def segmentation_to_US(
    segmentation: xr.DataArray, US_tracks: gpd.GeoDataFrame, tracker: str
) -> xr.Dataset | None:
    """


    Parameters
    ----------
    segmentation : xarray.core.dataarray.DataArray
        The output from the segmentation step of a given tracker.
    US_tracks : geopandas.geodataframe.GeoDataFrame
        US standard tracking output.
    tracker : str
        ["tobac"] The tracker used to generate the features.

    Raises
    ------
    Exception
        Exception if invalid tracker entered.

    Returns
    -------
    xarray.core.dataset.Dataset
        An xarray dataset matching the format of the CoCoMET-US segmentation specification.

    """

    if segmentation is None or US_tracks is None:
        return None

    if tracker.lower() == "tobac":
        feature_segmentation = (segmentation - 1).rename("Feature_Segmentation")
        cell_segmentation = deepcopy(feature_segmentation).rename("Cell_Segmentation")

        frame_groups = US_tracks.groupby("frame")

        # Loop over tracks, replacing feature_id values with cell_id values in the cell_segmenation DataArray
        for frame in tqdm(
            frame_groups,
            desc="=====Performing tobac Segmentation to US=====",
            total=frame_groups.ngroups,
        ):
            cell_segmentation_frame = cell_segmentation[frame[0]].values
            # Loop over each feature in that frame
            for feature in frame[1].itertuples():
                # Replace the feature_id with the cell_id
                cell_segmentation_frame[
                    cell_segmentation_frame == feature.feature_id
                ] = feature.cell_id

        # Combine into one xarray Dataset and return

        # To check for both prescense of altitude and shape of altitude without throwing DNE error
        altitude_check_bool = False
        if "altitude" in segmentation.coords:
            if segmentation.altitude.shape != ():
                altitude_check_bool = True

        if "model_level_number" in segmentation.coords:
            if segmentation.model_level_number.shape != ():
                altitude_check_bool = True

        # Check if altitude is present
        if altitude_check_bool:
            return_ds = xr.combine_by_coords([feature_segmentation, cell_segmentation])

            # Check for NEXRAD, and rename accordingly
            if (
                "y" in segmentation.dims
                and "x" in segmentation.dims
                and "z" in segmentation.dims
                and "lat" in segmentation.coords
                and "lon" in segmentation.coords
            ):
                return_ds = return_ds.assign_coords(
                    altitude=("z", feature_segmentation.z.values),
                    up_down=("z", np.arange(0, feature_segmentation.z.shape[0])),
                )
                return_ds = return_ds.swap_dims(
                    {"z": "up_down", "y": "south_north", "x": "west_east"}
                ).rename({"lat": "latitude", "lon": "longitude"})
                return_ds = return_ds.drop_vars(["z", "x", "y", "model_level_number"])

                return return_ds[
                    [
                        "time",
                        "up_down",
                        "south_north",
                        "west_east",
                        "Feature_Segmentation",
                        "Cell_Segmentation",
                    ]
                ]

            # For WRF case
            else:
                # Change altitude values to indices
                # return_ds = xr.combine_by_coords([feature_segmentation,cell_segmentation]).assign_coords({"up_down":np.arange(0,feature_segmentation.altitude.shape[0])})
                return_ds = return_ds.assign_coords(
                    up_down=(
                        "altitude",
                        np.arange(0, feature_segmentation.altitude.shape[0]),
                    )
                )

                # Remove extra coordinates and rename altitude
                return_ds = return_ds.swap_dims({"altitude": "up_down"}).drop_vars(
                    ["model_level_number", "x", "y"]
                )
                return return_ds[
                    [
                        "time",
                        "up_down",
                        "south_north",
                        "west_east",
                        "Feature_Segmentation",
                        "Cell_Segmentation",
                    ]
                ]

        else:
            # Concat into one dataset and remove superflous coordinates
            return_ds = xr.combine_by_coords([feature_segmentation, cell_segmentation])

            # Check for GOES and rename accordingly
            if "sensor_band_bit_depth" in segmentation.attrs:
                # Rename t to time and rename x and y values to south_north and west_east, respectively. Rename lat and lon to latitude and longitude
                return_ds = (
                    return_ds.assign_coords(time=("t", return_ds.t.values))
                    .swap_dims({"t": "time"})
                    .drop_vars(["t"])
                )
                return_ds = return_ds.swap_dims(
                    {"y": "south_north", "x": "west_east"}
                ).rename({"lat": "latitude", "lon": "longitude"})

                return_ds = return_ds.drop_vars(["x", "y"])
                return return_ds[
                    [
                        "time",
                        "south_north",
                        "west_east",
                        "Feature_Segmentation",
                        "Cell_Segmentation",
                    ]
                ]

            # Check for NEXRAD, and rename accordingly
            elif (
                "y" in segmentation.dims
                and "x" in segmentation.dims
                and "lat" in segmentation.coords
                and "lon" in segmentation.coords
            ):
                return_ds = return_ds.swap_dims(
                    {"y": "south_north", "x": "west_east"}
                ).rename({"lat": "latitude", "lon": "longitude"})

                return_ds = return_ds.drop_vars(["x", "y"])
                return return_ds[
                    [
                        "time",
                        "south_north",
                        "west_east",
                        "Feature_Segmentation",
                        "Cell_Segmentation",
                    ]
                ]

            # For WRF case
            else:
                return_ds = return_ds.drop_vars(["x", "y"])
                return return_ds[
                    [
                        "time",
                        "south_north",
                        "west_east",
                        "Feature_Segmentation",
                        "Cell_Segmentation",
                    ]
                ]

    raise Exception(f"!=====Invalid Tracker, You Entered: {tracker.lower()}=====!")


def bulk_moaap_to_US(
    mask: xr.Dataset,
    projection_x_coords: np.ndarray,
    projection_y_coords: np.ndarray,
    convert_type: str = "cloud",
) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame, xr.Dataset] | None:
    """


    Parameters
    ----------
    mask : xarray.core.dataset.Dataset
        An xarray file which is the default output from MOAAP and contains the mask information for all tracked types.
    projection_x_coords : np.ndarray
        Numpy array of projection x coordinates.
    projection_y_coords : np.ndarray
        Numpy array of projection y coordinates.
    convert_type : str, optional
        ["MCS", "Cloud"] The type of tracking data to extract. The default is "cloud".

    Raises
    ------
    Exception
        Exception if invalid phenomena type entered.

    Returns
    -------
    US_features : geopandas.geodataframe.GeoDataFrame
        Features geopandas dataframe following CoCoMET-US specification.
    US_linking : geopandas.geodataframe.GeoDataFrame
        Linking geopandas dataframe following CoCoMET-US specification.
    US_segmentation_2d : xarray.core.dataset.Dataset
        Segmentation xarray dataset following CoCoMET-US specification.

    """

    # Get dt
    dt = np.median(np.diff(mask.time).astype("timedelta64[m]")).astype(int)

    # Get the numpy array that contains the object we care about's mask
    if convert_type.lower() == "cloud":
        if "BT_Objects" not in mask:
            return None

        mask_field = mask.BT_Objects.values

    elif convert_type.lower() == "mcs":
        if "MCS_Tb_Objects" not in mask:
            return None

        mask_field = mask.MCS_Tb_Objects.values

    else:
        raise Exception(
            f"!=====Invalid Phenomena Type Selected. You Entered: {convert_type.lower()}!====="
        )

    segment_feature_mask = np.zeros(mask_field.shape) - 1
    segment_cell_mask = np.zeros(mask_field.shape) - 1

    frames = []
    times = []
    feature_ids = [-1]
    cell_ids = []
    lifetimes = []
    lifetime_percents = []
    south_norths = []
    west_easts = []
    lats = []
    lons = []
    proj_xs = []
    proj_ys = []

    lifetime_dict = {}

    # Loop over frames
    for ii in tqdm(
        range(mask_field.shape[0]),
        desc="=====Converting MOAAP to US=====",
        total=mask_field.shape[0],
    ):
        # Get unique cell ids for this frame
        unique_cells = np.unique(mask_field[ii])[1:]

        segment_cell_mask[ii] = mask_field[ii] - 1

        # Loop over unique cell ids
        for cell_id in unique_cells:
            segment_feature_mask[ii][mask_field[ii] == cell_id] = feature_ids[-1] + 1

            # Append timing information and cell/feature ids
            frames.append(ii)
            times.append(pd.Timestamp(mask.time.values[ii]))

            feature_ids.append(feature_ids[-1] + 1)
            cell_ids.append(int(cell_id) - 1)

            # If cell has not yet been tracked, add it to the lifetime_dict
            if int(cell_id) - 1 not in lifetime_dict:
                lifetime_dict[int(cell_id) - 1] = {
                    "frame": [ii],
                    "lifetime": [pd.Timedelta(0)],
                }
                lifetimes.append(pd.Timedelta(0))

            else:
                new_time = lifetime_dict[int(cell_id) - 1]["lifetime"][-1] + (
                    pd.Timedelta(minutes=dt)
                    * (ii - lifetime_dict[int(cell_id) - 1]["frame"][-1])
                )
                lifetime_dict[int(cell_id) - 1]["lifetime"].append(new_time)
                lifetime_dict[int(cell_id) - 1]["frame"].append(ii)
                lifetimes.append(new_time)

            # Calculate cell centroid
            cell_mask = mask_field[ii] == cell_id
            centroid = center_of_mass(cell_mask)

            south_norths.append(centroid[0])
            west_easts.append(centroid[1])

    temp_linking = pd.DataFrame(data={"lifetime": lifetimes, "cell_id": cell_ids})

    # Loop over rows to calculate lifetime percents
    for row in tqdm(
        temp_linking.iterrows(),
        desc="=====Processing MOAAP Cell Lifetimes====",
        total=temp_linking.shape[0],
    ):
        cell_max_life = temp_linking.query(
            "cell_id==@row[1].cell_id"
        ).lifetime.values.max()

        # If only tracked one time, add -1 to lifetime_percent
        if cell_max_life == 0:
            lifetime_percents.append(-1)
        else:
            lifetime_percents.append(row[1].lifetime / cell_max_life)

    # Calculate lat and lons by interpolating over lat/lon fields
    lat_dims = (np.arange(0, mask.lat[0].shape[0]), np.arange(0, mask.lat[0].shape[1]))
    lon_dims = (np.arange(0, mask.lon[0].shape[0]), np.arange(0, mask.lon[0].shape[1]))

    lat = interpn(lat_dims, mask.lat[0].values, np.array([south_norths, west_easts]).T)
    lon = interpn(lon_dims, mask.lon[0].values, np.array([south_norths, west_easts]).T)

    lats.extend(lat)
    lons.extend(lon)

    # Calculate projection x and y coordinates via interpolating again
    proj_x_coords = np.arange(0, mask.lat[0].shape[1])
    proj_y_coords = np.arange(0, mask.lat[0].shape[0])

    new_proj_x = interpn([proj_x_coords], projection_x_coords, west_easts)
    new_proj_y = interpn([proj_y_coords], projection_y_coords, south_norths)

    proj_xs.extend(new_proj_x)
    proj_ys.extend(new_proj_y)

    # Create all dataframs and convert them to geodataframes
    US_features = pd.DataFrame(
        data={
            "frame": frames,
            "time": times,
            "feature_id": feature_ids[1:],
            "south_north": south_norths,
            "west_east": west_easts,
            "up_down": np.repeat(np.nan, len(frames)),
            "latitude": lats,
            "longitude": lons,
            "projection_x": new_proj_x,
            "projection_y": new_proj_y,
            "altitude": np.repeat(np.nan, len(frames)),
        }
    )

    US_features = gpd.GeoDataFrame(
        US_features,
        geometry=gpd.points_from_xy(US_features.longitude, US_features.latitude),
        crs="EPSG:4326",
    )

    US_linking = pd.DataFrame(
        data={
            "frame": frames,
            "time": times,
            "lifetime": lifetimes,
            "lifetime_percent": lifetime_percents,
            "feature_id": feature_ids[1:],
            "cell_id": cell_ids,
            "south_north": south_norths,
            "west_east": west_easts,
            "up_down": np.repeat(np.nan, len(frames)),
            "latitude": lats,
            "longitude": lons,
            "projection_x": new_proj_x,
            "projection_y": new_proj_y,
            "altitude": np.repeat(np.nan, len(frames)),
        }
    )

    US_linking = gpd.GeoDataFrame(
        US_linking,
        geometry=gpd.points_from_xy(US_linking.longitude, US_linking.latitude),
        crs="EPSG:4326",
    )

    # Now perform segmentation conversion
    US_segmentation_2d = xr.Dataset(
        coords=dict(
            time=mask.time.values,
            south_north=mask.yc.values,
            west_east=mask.xc.values,
            projection_y_coordinate=("south_north", projection_y_coords),
            projection_x_coordinate=("west_east", projection_x_coords),
            latitude=(["south_north", "west_east"], mask.lat[0].values),
            longitude=(["south_north", "west_east"], mask.lon[0].values),
        ),
        data_vars=dict(
            Feature_Segmentation=(
                ["time", "south_north", "west_east"],
                segment_feature_mask,
            ),
            Cell_Segmentation=(["time", "south_north", "west_east"], segment_cell_mask),
        ),
        attrs=dict(description="Mask for tracked objects from MOAAP"),
    )

    return (US_features, US_linking, US_segmentation_2d)


def bulk_tams_to_US(
    output: gpd.GeoDataFrame,
    latlon_coord_system: tuple[np.ndarray, np.ndarray],
    projection_x_coords: np.ndarray,
    projection_y_coords: np.ndarray,
    convert_type: str = "cloud",
) -> tuple[gpd.GeoDataFrame, gpd.GeoDataFrame, xr.Dataset] | None:
    """


    Parameters
    ----------
    output : pd.DataFrame
        A geodataframe with geometry of tracked cloud objects
    latlon_coord_system : tuple[np.ndarray, np.ndarray]
        A tuple of the latitude and longitude coordinate numpy arrays
    projection_x_coords : np.ndarray
        A numpy array of the x projection coordinates.
    projection_y_coords : np.ndarray
        A numpy array of the y projection coordinates.
    convert_type : str, optional
        ["MCS", "Cloud"] The type of tracking data to extract. The default is "cloud".

    Raises
    ------
    Exception
        Exception if invalid phenomena type entered.

    Returns
    -------
    US_features : geopandas.geodataframe.GeoDataFrame
        Features geopandas dataframe following CoCoMET-US specification.
    US_linking : geopandas.geodataframe.GeoDataFrame
        Linking geopandas dataframe following CoCoMET-US specification.
    US_segmentation_2d : xarray.core.dataset.Dataset
        Segmentation xarray dataset following CoCoMET-US specification.
    """

    # Get the numpy array that contains the object we care about's mask
    if convert_type.lower() == "cloud":
        tracked_elements = output

    # TODO : fix this once you get gridding data aggregation working for TAMS. Also make it output[0] above
    # elif convert_type.lower() == "mcs":

    #     tracked_elements = output[1]

    else:
        raise Exception(
            f"!=====Invalid Phenomena Type Selected. You Entered: {convert_type.lower()}!====="
        )

    # Get dt
    dt = np.median(tracked_elements["dtime"].dt.total_seconds().astype(int))
    cell_mask = convert_df_to_mask(tracked_elements, latlon_coord_system)
    feature_mask = convert_cell_mask_to_feature_mask(cell_mask)

    frames = []
    times = []
    feature_ids = [-1]
    cell_ids = []
    lifetimes = []
    lifetime_percents = []
    south_norths = []
    west_easts = []
    lats = []
    lons = []
    new_proj_x = []
    new_proj_y = []

    lifetime_dict = {}

    # Iterate through each timestep
    for ii in tqdm(
        range(len(tracked_elements)),
        desc="=====Converting TAMS to US=====",
        total=len(tracked_elements),
    ):
        frame = tracked_elements["itime"][ii]
        frames.append(frame)
        times.append(tracked_elements["time"][ii])

        cell_id = int(tracked_elements["mcs_id"][ii])
        cell_ids.append(cell_id)

        lat = tracked_elements["geometry"][ii].centroid.y
        lon = tracked_elements["geometry"][ii].centroid.x
        lats.append(lat)
        lons.append(lon)

        latcoords = latlon_coord_system[0][frames[-1]].values
        loncoords = latlon_coord_system[1][frames[-1]].values

        # TODO: find a way to interpolate the grid coordinates from the latitude and longitude arrays
        # currently, _find_closest_latlon snaps to the closest grid coordinate for a given latitude and longitude coordiante
        closest_point = _find_closest_latlon(lat, lon, latcoords, loncoords)

        feature_ids.append(feature_mask[frame, *closest_point])

        south_norths.append(closest_point[0])
        west_easts.append(closest_point[1])

        new_proj_x.append(projection_x_coords[west_easts[-1]])
        new_proj_y.append(projection_y_coords[south_norths[-1]])

        # If cell has not yet been tracked, add it to the lifetime_dict
        if int(cell_id) not in lifetime_dict:
            lifetime_dict[int(cell_id)] = {
                "frame": [frame],
                "lifetime": [pd.Timedelta(0)],
            }
            lifetimes.append(pd.Timedelta(0))

        else:
            new_time = lifetime_dict[int(cell_id)]["lifetime"][-1] + (
                pd.Timedelta(minutes=dt)
                * (frame - lifetime_dict[int(cell_id)]["frame"][-1])
            )
            lifetime_dict[int(cell_id)]["lifetime"].append(new_time)
            lifetime_dict[int(cell_id)]["frame"].append(frame)
            lifetimes.append(new_time)

    temp_linking = pd.DataFrame(data={"lifetime": lifetimes, "cell_id": cell_ids})

    # Loop over rows to calculate lifetime percents
    for row in tqdm(
        temp_linking.iterrows(),
        desc="=====Processing TAMS Cell Lifetimes====",
        total=temp_linking.shape[0],
    ):
        cell_max_life = temp_linking.query(
            "cell_id==@row[1].cell_id"
        ).lifetime.values.max()

        # If only tracked one time, add -1 to lifetime_percent
        if cell_max_life == 0:
            lifetime_percents.append(-1)
        else:
            lifetime_percents.append(row[1].lifetime / cell_max_life)

    # Create all dataframs and convert them to geodataframes
    US_features = pd.DataFrame(
        data={
            "frame": frames,
            "time": times,
            "feature_id": feature_ids[1:],
            "south_north": south_norths,
            "west_east": west_easts,
            "up_down": np.repeat(np.nan, len(frames)),
            "latitude": lats,
            "longitude": lons,
            "projection_x": new_proj_x,
            "projection_y": new_proj_y,
            "altitude": np.repeat(np.nan, len(frames)),
        }
    )

    US_features = gpd.GeoDataFrame(
        US_features,
        geometry=gpd.points_from_xy(US_features.longitude, US_features.latitude),
        crs="EPSG:4326",
    )

    US_linking = pd.DataFrame(
        data={
            "frame": frames,
            "time": times,
            "lifetime": lifetimes,
            "lifetime_percent": lifetime_percents,
            "feature_id": feature_ids[1:],
            "cell_id": cell_ids,
            "south_north": south_norths,
            "west_east": west_easts,
            "up_down": np.repeat(np.nan, len(frames)),
            "latitude": lats,
            "longitude": lons,
            "projection_x": new_proj_x,
            "projection_y": new_proj_y,
            "altitude": np.repeat(np.nan, len(frames)),
        }
    )

    US_linking = gpd.GeoDataFrame(
        US_linking,
        geometry=gpd.points_from_xy(US_linking.longitude, US_linking.latitude),
        crs="EPSG:4326",
    )

    # Now perform segmentation conversion
    US_segmentation_2d = xr.Dataset(
        coords=dict(
            time=np.arange(latlon_coord_system[0].shape[0]),
            south_north=np.arange(latlon_coord_system[0].shape[1]),
            west_east=np.arange(latlon_coord_system[0].shape[2]),
            projection_y_coordinate=("south_north", projection_y_coords),
            projection_x_coordinate=("west_east", projection_x_coords),
            latitude=(["south_north", "west_east"], latlon_coord_system[0][0].values),
            longitude=(["south_north", "west_east"], latlon_coord_system[1][0].values),
        ),
        data_vars=dict(
            Feature_Segmentation=(
                ["time", "south_north", "west_east"],
                convert_cell_mask_to_feature_mask(cell_mask),
            ),
            Cell_Segmentation=(["time", "south_north", "west_east"], cell_mask),
        ),
        attrs=dict(description="Mask for tracked objects from TAMS"),
    )

    return (US_features, US_linking, US_segmentation_2d)


##################################################################################
##################################################################################
# Extra functions for converting TAMS dataframe into a mask


def _find_closest_latlon(lat_true, lon_true, lat_arr, lon_arr):
    # TODO : find a more exact way to do this - get an x, y position in grid coordinates
    #   given a latitude, a longitude, an array of latitudes, and an array of longitudes
    diff_arr_lat = np.abs(lat_arr - lat_true)
    diff_arr_lon = np.abs(lon_arr - lon_true)
    squared_diff_arr = diff_arr_lat**2 + diff_arr_lon**2
    closest_point = np.unravel_index(squared_diff_arr.argmin(), squared_diff_arr.shape)
    # return (closest_point[1], closest_point[0])
    return closest_point


def _points_from_polygons(polygons):
    from shapely.geometry import MultiPolygon

    points = []
    for mpoly in polygons:
        if isinstance(mpoly, MultiPolygon):
            polys = list(mpoly)
        else:
            polys = [mpoly]
        for polygon in polys:
            for point in polygon.exterior.coords:
                points.append(point)
            for interior in polygon.interiors:
                for point in interior.coords:
                    points.append(point)
    return points


def _return_gridpoints(points, latlon_coord_system):
    gridpoints = []
    latcoords = latlon_coord_system[0].values[0]
    loncoords = latlon_coord_system[1].values[0]
    for ilon, ilat in points:
        x, y = _find_closest_latlon(ilat, ilon, latcoords, loncoords)
        gridpoints.append((y, x))

    return gridpoints


def convert_df_to_mask(ce, latlon_coord_system):
    full_mask = np.zeros_like(latlon_coord_system[0].values)

    # TODO : this is really slow, find a better way to do this
    for frame in tqdm(np.unique(ce.itime),
                    desc="=====Converting TAMS Dataframe to Mask=====",
                    total = len(np.unique(ce.itime))):
        id_groups = ce[ce.itime == frame].groupby("mcs_id")

        for mcs_id, grouped_df in id_groups:
            points = _points_from_polygons(
                grouped_df.geometry
            )  # get the points in lat, lon
            gridpoints = _return_gridpoints(
                points, latlon_coord_system
            )  # convert the points to gridspace

            ny, nx = latlon_coord_system[0].shape[1:]
            poly_verts = gridpoints

            x, y = np.meshgrid(np.arange(nx), np.arange(ny))
            x, y = x.flatten(), y.flatten()

            points = np.vstack((x, y)).T

            path = Path(poly_verts)
            grid = path.contains_points(points)
            grid = grid.reshape((ny, nx))

            grid = np.array(grid, dtype=int) * (mcs_id + 1)

            non_zero_grid = grid != 0
            full_mask_in_frame = full_mask[frame]
            full_mask_in_frame[non_zero_grid] = grid[non_zero_grid]
    full_mask -= 1
    return full_mask


def convert_cell_mask_to_feature_mask(cell_mask):
    feature_mask = deepcopy(cell_mask)
    last_feature = 0
    for t in range(feature_mask.shape[0]):
        real_cell_bool = feature_mask[t] > -1
        feature_mask[t][real_cell_bool] += last_feature
        last_feature = np.max(feature_mask[t])

    return feature_mask
